"""
"""
from __future__ import annotations

import os
import shutil
import abipy.core.abinit_units as abu
from pathlib import Path
from ase.calculators.singlepoint import SinglePointCalculator
from monty.string import list_strings  # marquee,
from monty.termcolor import cprint
from ase.io import read
from ase.stress import full_3x3_to_voigt_6_stress
from ase.io import write
from pymatgen.io.vasp.outputs import Vasprun, Outcar
from pymatgen.io.vasp.sets import MatPESStaticSet # , MPStaticSet
from abipy.core import Structure
from abipy.electrons.gsr import GsrFile
from abipy.tools.typing import PathLike
import abipy.flowtk.qutils as qu
from abipy.ml.tools import get_energy_step


class ExtxyzIOWriter:
    """
    Example:

        # To find all the vasprun.xml files starting from the a top-level directory, use:

            xyz_writer = ExtxyzIOWriter.from_top(".", "vasprun.xml")
            print(xyz_writer)
            xyz_writer.write("foo.xyz")

        # A similar syntax can be used for Abinit GSR files:

            ExtxyzIOWriter.from_top(".", "GSR.nc").write("foo.xyz")

        # To specify an explicit list of files, use:

            xyz_writer = ExtxyzIOWriter(["dir1/vasprun.xml", "dir2/vasprun.xml"])
            xyz_writer.write("foo.xyz")
    """

    SUPPORTED_EXTS = [
        "vasprun.xml",
        "GSR.nc",
    ]

    @classmethod
    def from_top(cls, top: PathLike, ext: str):
        """
        Scan for files with extension ext starting from the top directory top.
        """
        from monty.os.path import find_exts
        filepaths = find_exts(str(top), ext)
        return cls(filepaths)

    def __init__(self, filepaths: list[PathLike]):
        self.filepaths = list_strings(filepaths)
        if not self.filepaths:
            raise RuntimeError("Empty list of filepaths!")

        for ext in self.SUPPORTED_EXTS:
            if all(f.endswith(ext) for f in self.filepaths):
                self.ext = ext
                break
        else:
            raise ValueError(f"Cannot detect extension from filepaths, should be in: {self.SUPPORTED_EXTS}")

    def to_string(self, verbose: int = 0) -> str:
        """String representation with verbosiy level ``verbose``."""
        lines = []
        for i, path in enumerate(self.filepaths):
            lines.append(f"[{i}]: {path}")
        return "\n".join(lines)

    def __str__(self) -> str:
        return self.to_string()

    def write(self, xyz_filepath: PathLike, overwrite: bool = False):
        """
        """
        if not overwrite and os.path.isfile(xyz_filepath):
            raise RuntimeError(f"Cannot overwrite pre-existent file: {xyz_filepath=}, use overwrite=True to allow overwriting.")

        with open(xyz_filepath, "wt") as fh:
            for atoms in self.yield_atoms():
                write(fh, atoms, format='extxyz', append=True)

    def yield_atoms(self):
        """
        """
        for filepath in self.filepaths:
            if self.ext == "vasprun.xml":
                vasprun = Vasprun(filepath)
                dirname = os.path.dirname(filepath)
                outcar_path = os.path.join(dirname, "OUTCAR")
                outcar = Outcar(outcar_path) if os.path.exists(outcar_path) else None

                ok = check_vasp_success(vasprun, outcar, verbose=1)

                last_step = vasprun.ionic_steps[-1]
                structure, forces, stress = last_step["structure"], last_step["forces"], last_step["stress"]
                energy = get_energy_step(last_step)

            elif self.ext == "GSR.nc":
                with GsrFile(filepath) as gsr:
                    if not gsr.is_scf_run:
                        raise RuntimeError("GSR file was not produced by a SCF run!")
                    structure, forces, stress_gpa = gsr.structure, gsr.cart_forces, gsr.cart_stress_tensor
                    stress = stress_gpa / abu.eVA3_GPa
                    energy = float(gsr.energy)

            else:
                raise ValueError(f"Format {self.ext=} is not supported!")

            atoms = structure.to_ase_atoms()
            stress = full_3x3_to_voigt_6_stress(stress)

            # Attach calculator with results.
            atoms.calc = SinglePointCalculator(atoms,
                                               energy=energy,
                                               free_energy=energy,
                                               forces=forces,
                                               stress=stress,
                                               )
            yield atoms


def check_vasp_success(vasprun: Vasprun, outcar: Outcar | None, verbose: int = 1) -> bool:
    """
    Check if a VASP calculation completed successfully.

    Returns: True if the calculation completed successfully, False otherwise.
    """
    def my_print(*args, **kwargs):
        if verbose: print(*args, **kwargs)

    try:
        if not vasprun.converged:
            my_print("Calculation did not converge.")
            return False

        if outcar is not None:
            if outcar.run_stats.get("Elapsed time (sec)"):
                my_print("Calculation completed in {} seconds.".format(outcar.run_stats["Elapsed time (sec)"]))
            else:
                my_print("Elapsed time not found in OUTCAR.")
                return False

        my_print("Calculation completed successfully.")
        return True

    except Exception as e:
        my_print(f"Error checking calculation status: {e}")
        return False


class SinglePointRunner:
    """

    Usage example:

    .. code-block:: python

        traj_range = range(0, -1, 100)
        runner = SinglePointRunner("out.traj", "outdir", traj_range)
        runner.sbatch()
        runner.collect_xyz("foo.xyz")
    """
    slurm_script_name = "run.sh"

    custodian_script_name = "run_custodian.py"

    def __init__(self, traj_path: PathLike, traj_range: range,
                 topdir: PathLike = ".", code: str = "vasp",
                 vasp_set_cls=MatPESStaticSet,
                 verbose: int = 0):
        """
        Args:
            traj_path: Path to ASE trajectory file.
            traj_range:
            topdir:
            code:
            verbose:
        """
        self.traj_path = traj_path
        self.topdir = Path(str(topdir)).absolute()
        self.traj_range = traj_range
        if not isinstance(traj_range, range):
            raise TypeError(f"Got {type(traj_range)} instead of range")
        self.code = code

        err_lines = []
        slurm_body = ""

        if code == "vasp":
            self.vasp_set_cls = vasp_set_cls
            slurm_body = f"python {self.custodian_script_name}"
            if not os.path.exists(self.custodian_script_name):
                open(self.custodian_script_name, "wt").write(qu.get_custodian_template())
                err_lines.append(f"""\
No custodian script: {self.custodian_script_name} has been found in {str(self.topdir)}.
A template that requires customization has been generated for you!""")
            else:
                self.custodian_script_str = open(self.custodian_script_name, "rt").read()

        if not os.path.exists(self.slurm_script_name):
            open(self.slurm_script_name, "wt").write(qu.get_slurm_template(slurm_body))
            err_lines.append(f"""\
No slurm submission script: {self.slurm_script_name} has been found in {str(self.topdir)}.
A template that requires customization has been generated for you!""")
        else:
            self.slurm_script_str = open(self.slurm_script_name, "rt").read()

        if err_lines:
            raise RuntimeError("\n".join(err_lines))

        self.verbose = int(verbose)

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation with verbosiy level ``verbose``."""
        lines = []
        app = lines.append

        return "\n".join(lines)

    def sbatch(self, max_jobs: int = 100) -> list[int]:
        """
        Submit max_jobs SinglePoint calculations with structures taken from the ASE trajectory file.
        """
        if not self.topdir.exists(): self.topdir.mkdir()

        job_ids = []
        for index in self.traj_range:
            workdir = self.topdir / f"SINGLEPOINT_{index}"
            if workdir.exists():
                print(f"{str(workdir)} already exists. Ignoring it")
                continue

            try:
                atoms = read(self.traj_path, index=index)
            except StopIteration as exc:
                print(f"ASE trajectory does not have more that {index+1} configurations. Exiting sbatch loop!")
                break

            structure = Structure.as_structure(atoms)
            workdir.mkdir()

            if self.code == "vasp":
                # Generate VASP input files using the Materials Project settings for a single-point calculation

                user_incar_settings = {
                    "NCORE": 2,
                    'LWAVE': False,       # Do not write WAVECAR
                    'LCHARG': False,      # Do not Write CHGCAR
                }
                vasp_input_set = self.vasp_set_cls(structure, user_incar_settings=user_incar_settings)
                vasp_input_set.write_input(workdir)
                with open(workdir / self.custodian_script_name, "wt") as fh:
                    fh.write(self.custodian_script_str)

            else:
                raise ValueError(f"Unsupported {self.code=}")

            try:
                job_id = qu.slurm_write_and_sbatch(workdir / "run.sh", self.slurm_script_str)

            except Exception as exc:
                cprint(exc, "red")
                cprint("Job sumbission failed. Will remove directory and exit sbatch loop.", color="red")
                shutil.rmtree(workdir)
                break

            job_ids.append(job_id)
            if len(job_ids) == max_jobs:
                print(f"Reached {max_jobs=}, will stop firing new jobs!")

        return job_ids

    def write_xyz(self, xyz_filepath: PathLike, dry_run=False) -> None:
        """
        """
        ext = {
            "vasp": "vasprun.xml",
            "abinit": "GSR.nc",
        }[self.code]

        writer = ExtxyzIOWriter.from_top(self.topdir, ext)
        writer.write(xyz_filepath)



#!/usr/bin/env python
r"""
G0W0 convergence study
======================

G0W0 convergence study wrt ecuteps and the number of bands in W.
"""
from __future__ import print_function, division, unicode_literals, absolute_import

import sys
import os
import numpy as np

import abipy.abilab as abilab
import abipy.data as abidata
import abipy.flowtk as flowtk

# system aopt SR DFT FR DFT SR G0W0 FR G0W0 aexpt157 SR DFT SR G0W0 expt.157
alatt = {
"Al-Sb": 6.23, # 1.22 1.02 1.89 1.65 6.136 1.20 2.23 1.60
"Cd-S": 5.93,  # 1.02 1.01 1.61 1.59 5.832 1.14 1.77 2.42
"Cd-Te": 6.63, # 0.56 0.29 1.07 0.76 6.477 0.76 1.29 1.44
"Ga-As": 5.75, # 0.15 0.01 0.25 0.13 5.653 0.50 0.62 1.35
"Ga-P": 5.51,  # 1.54 1.51 1.95 1.91 5.450 1.63 2.17 2.24
"In-P": 5.96,  # 0.42 0.39 0.61 0.57 5.869 0.68 0.90 1.27
"Si-Si": 5.48,   # 0.72 0.71 1.38 1.36 5.431 0.67 1.29 1.12
"Zn-S": 5.45,  # 1.99 1.97 3.08 3.06 5.409 2.09 3.26 3.54
"Zn-Se": 5.73, # 1.13 1.01 1.89 1.74 5.668 1.26 2.06 2.58
"Zn-Te": 6.18, # 1.08 0.79 1.87 1.49 6.101 1.24 2.08 2.26
}


def make_inputs(nspinor, paral_kgb=1):
    """
    Returns a tuple of 4 input files for SCF, NSCF, SCR, SIGMA calculations.
    These files are then used as templates for the convergence study
    wrt ecuteps and the number of bands in W.
    """
    formula = "Si-Si"
    species = formula.split("-")
    assert len(species) == 2
    a = alatt[formula]
    pseudos = "Si_r.psp8"
    #pseudos = abilab.PseudoTable.from_dir()
    structure = abilab.Structure.fcc(a, species)
    structure = abidata.structure_from_ucell("Si")
    multi = abilab.MultiDataset(structure, pseudos, ndtset=5)
    #pseudos = multi[0].pseudos

    #kptopt=1 if nspinor == 1 else 4
    kptopt=4
    ngkpt = [4, 4, 4]
    shiftk = [0, 0, 0]

    ecut = 4
    ecuteps = 4
    nband_gs = 6 * nspinor
    nband_mbpt = 10 * nspinor
    nbdbuf = 8

    # non-collinear GS (activate SOC allow for magnetic solution although Si is not)
    global_vars = dict(
        nspinor=nspinor,
        nspden=1 if nspinor == 1 else 4,
        so_psp="*0" if nspinor == 1 else "*1",
        ecut=ecut,
        istwfk="*1",
        paral_kgb=paral_kgb,
        #nsym=1,
        #iomode=1,
    )

    multi.set_vars(global_vars)
    multi.set_kmesh(ngkpt=ngkpt, shiftk=shiftk, kptopt=kptopt)

    # SCF
    multi[0].set_vars(
        nband=nband_gs,
        tolvrs=1.e-10,
    )

    # NSCF on k-path
    multi[1].set_vars(
        nband=nband_gs + nbdbuf,
        nbdbuf=nbdbuf,
        tolwfr=1.e-18,
        iscf=-2,
    )
    multi[1].set_kpath(ndivsm=10)

    # NSCF on k-mesh
    multi[2].set_vars(
        nband=nband_gs + nbdbuf,
        nbdbuf=nbdbuf,
        tolwfr=1.e-18,
        iscf=-2
    )

    # SCR
    multi[3].set_vars(
        optdriver=3,
        ecutwfn=ecut,
        ecuteps=ecuteps,
        nband=nband_mbpt,
        symchi=1,
        awtr=1 if kptopt == 1 else 0,
        gwpara=2 if kptopt == 1 else 1,
        inclvkb=0,
    )

    # SIGMA
    multi[4].set_vars(
        optdriver=4,
        nband=nband_mbpt,
        ecutwfn=ecut,
        ecuteps=ecuteps,
        ecutsigx=ecut,
        #ecutsigx=(4*ecut), ! This is problematic
        symsigma=1,
        gw_qprange=0,
        )

    #multi[4].set_kptgw(kptgw=[[0,0,0], [0.5, 0, 0]], bdgw=[1, 8])

    return multi.split_datasets()


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    flow = flowtk.Flow(workdir=options.workdir, manager=options.manager)

    for nspinor in (1, 2):
    #for nspinor in (2,):
        # Get our templates
        scf_inp, bands_inp, nscf_inp, scr_inp, sig_inp = make_inputs(nspinor)

        # Band structure work to produce the WFK file
        bands_work = flowtk.BandStructureWork(scf_inp, bands_inp, dos_inputs=[nscf_inp])
        flow.register_work(bands_work)

        # Build a work made of two SCR runs with different value of nband
        gw_work = flowtk.Work()
        scr_task = gw_work.register_scr_task(scr_inp, deps={bands_work[2]: "WFK"})
        gw_work.register_sigma_task(sig_inp, deps={bands_work[2]: "WFK", scr_task: "SCR"})
        flow.register_work(gw_work)

    return flow


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__=="__main__":
    sys.exit(main())



#!/usr/bin/env python

#--- !WARNING
s = """
---
src_file: mpi_setup.F90
src_line: 692
message: |
    The number of KPT processors, npkpt, should be a multiple of nkpt*nsppol.
    However, npkpt=2 and nkpt*nsppol=97
...
"""

from ruamel.yaml import YAML

yaml = YAML(typ='safe')   # default, if not specfied, is 'rt' (round-trip)
doc = yaml.load(s)

print(doc)



#!/usr/bin/env python
r"""
G0W0 flow with convergence study wrt nband
==========================================

This script shows how to compute the G0W0 corrections in silicon.
More specifically, we build a flow to analyze the convergence of the QP corrections
wrt to the number of bands in the self-energy. More complicated convergence studies
can be implemented on the basis of this example.
"""

import os
import sys
import abipy.data as data
import abipy.abilab as abilab

from abipy import flowtk


def make_inputs(ngkpt, paral_kgb=1):
    # Crystalline silicon
    # Calculation of the GW correction to the direct band gap in Gamma
    # Dataset 1: ground state calculation
    # Dataset 2: NSCF calculation
    # Dataset 3: calculation of the screening
    # Dataset 4-5-6: Self-Energy matrix elements (GW corrections) with different values of nband

    multi = abilab.MultiDataset(structure=data.cif_file("si.cif"),
                                pseudos=data.pseudos("14si.pspnc"), ndtset=6)

    # This grid is the most economical, but does not contain the Gamma point.
    scf_kmesh = dict(
        ngkpt=ngkpt,
        shiftk=[0.5, 0.5, 0.5,
                0.5, 0.0, 0.0,
                0.0, 0.5, 0.0,
                0.0, 0.0, 0.5]
    )

    # This grid contains the Gamma point, which is the point at which
    # we will compute the (direct) band gap.
    gw_kmesh = dict(
        ngkpt=ngkpt,
        shiftk=[0.0, 0.0, 0.0,
                0.0, 0.5, 0.5,
                0.5, 0.0, 0.5,
                0.5, 0.5, 0.0]
    )

    # Global variables. gw_kmesh is used in all datasets except DATASET 1.
    ecut = 6

    multi.set_vars(
        ecut=ecut,
        timopt=-1,
        istwfk="*1",
        paral_kgb=paral_kgb,
        gwpara=2,
    )
    multi.set_kmesh(**gw_kmesh)

    # Dataset 1 (GS run)
    multi[0].set_kmesh(**scf_kmesh)
    multi[0].set_vars(
        tolvrs=1e-6,
        nband=4,
    )

    # Dataset 2 (NSCF run)
    # Here we select the second dataset directly with the syntax multi[1]
    multi[1].set_vars(
        iscf=-2,
        tolwfr=1e-12,
        nband=35,
        nbdbuf=5,
    )

    # Dataset3: Calculation of the screening.
    multi[2].set_vars(
        optdriver=3,
        nband=25,
        ecutwfn=ecut,
        symchi=1,
        inclvkb=0,
        ecuteps=4.0,
        ppmfrq="16.7 eV",
    )

    # Dataset4: Calculation of the Self-Energy matrix elements (GW corrections)
    kptgw = [
         -2.50000000E-01, -2.50000000E-01,  0.00000000E+00,
         -2.50000000E-01,  2.50000000E-01,  0.00000000E+00,
          5.00000000E-01,  5.00000000E-01,  0.00000000E+00,
         -2.50000000E-01,  5.00000000E-01,  2.50000000E-01,
          5.00000000E-01,  0.00000000E+00,  0.00000000E+00,
          0.00000000E+00,  0.00000000E+00,  0.00000000E+00,
      ]

    bdgw = [1, 8]

    # Convergence study wrt nband in sigma.
    for idx, nband in enumerate([10, 20, 30]):
        multi[3+idx].set_vars(
            optdriver=4,
            nband=nband,
            ecutwfn=ecut,
            ecuteps=4.0,
            ecutsigx=6.0,
            symsigma=1,
            #gw_qprange=0,
            #nkptgw=0,
        )
        multi[3+idx].set_kptgw(kptgw, bdgw)

    return multi.split_datasets()


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_","flow_")

    # Change the value of ngkpt below to perform a GW calculation with a different k-mesh.
    scf, nscf, scr, sig1, sig2, sig3 = make_inputs(ngkpt=[2, 2, 2])

    return flowtk.g0w0_flow(options.workdir, scf, nscf, scr, [sig1, sig2, sig3], manager=options.manager)


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_si_g0w0.py -s
#
# The last three tasks (``w0_t3``, ``w0_t4``, ``w0_t5``) are the SigmaTask who have produced
# a netcdf file with the GW results with different number of bands.
# We can check this with the command:
#
#    abirun.py flow_si_g0w0/ listext SIGRES
#
# .. code-block:: bash
#
#       Found 3 files with extension `SIGRES` produced by the flow
#       File                                        Size [Mb]    Node_ID  Node Class
#       ----------------------------------------  -----------  ---------  ------------
#       flow_si_g0w0/w0/t3/outdata/out_SIGRES.nc         0.05     241325  SigmaTask
#       flow_si_g0w0/w0/t4/outdata/out_SIGRES.nc         0.08     241326  SigmaTask
#       flow_si_g0w0/w0/t5/outdata/out_SIGRES.nc         0.13     241327  SigmaTask
#
# Let's use the SIGRES robot to collect and analyze the results:
#
#    abirun.py flow_si_g0w0/ robot SIGRES
#
# and then, inside the ipython terminal, type:
#
# .. code-block:: ipython
#
#       In [1]: df = robot.get_dataframe()
#       In [2]: df
#       Out[2]:
#                                                 nsppol     qpgap            ecutwfn  \
#       flow_si_g0w0/w0/t3/outdata/out_SIGRES.nc       1  3.627960  5.914381651684836
#       flow_si_g0w0/w0/t4/outdata/out_SIGRES.nc       1  3.531781  5.914381651684836
#       flow_si_g0w0/w0/t5/outdata/out_SIGRES.nc       1  3.512285  5.914381651684836
#
#                                                            ecuteps  \
#       flow_si_g0w0/w0/t3/outdata/out_SIGRES.nc  3.6964885323070074
#       flow_si_g0w0/w0/t4/outdata/out_SIGRES.nc  3.6964885323070074
#       flow_si_g0w0/w0/t5/outdata/out_SIGRES.nc  3.6964885323070074
#
#                                                          ecutsigx scr_nband  \
#       flow_si_g0w0/w0/t3/outdata/out_SIGRES.nc  5.914381651684846        25
#       flow_si_g0w0/w0/t4/outdata/out_SIGRES.nc  5.914381651684846        25
#       flow_si_g0w0/w0/t5/outdata/out_SIGRES.nc  5.914381651684846        25
#
#                                                sigma_nband gwcalctyp scissor_ene  \
#       flow_si_g0w0/w0/t3/outdata/out_SIGRES.nc          10         0         0.0
#       flow_si_g0w0/w0/t4/outdata/out_SIGRES.nc          20         0         0.0
#       flow_si_g0w0/w0/t5/outdata/out_SIGRES.nc          30         0         0.0
#
#                                                 nkibz
#       flow_si_g0w0/w0/t3/outdata/out_SIGRES.nc      6
#       flow_si_g0w0/w0/t4/outdata/out_SIGRES.nc      6
#       flow_si_g0w0/w0/t5/outdata/out_SIGRES.nc      6
#
#       In [3]: %matplotlib
#       In [4]: df.plot("sigma_nband", "qpgap", marker="o")
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_si_g0w0.png?raw=true
#    :alt: QP results in Si plotted vs the KS energy e0.
#



#!/usr/bin/env python
"""Benchmark for phonon calculation with DFPT."""
import sys
import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata

from itertools import product
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_inputs(paw=False):
    """
    This function constructs the input files for the phonon calculation:
    GS input + the input files for the phonon calculation.
    """
    # Crystalline AlAs: computation of the second derivative of the total energy
    structure = abidata.structure_from_ucell("AlAs")
    pseudos = abidata.pseudos("13al.981214.fhi", "33as.pspnc")
    if paw:
        raise NotImplementedError("PAW")

    # List of q-points for the phonon calculation.
    qpoints = [
             0.00000000E+00,  0.00000000E+00,  0.00000000E+00,
             2.50000000E-01,  0.00000000E+00,  0.00000000E+00,
             5.00000000E-01,  0.00000000E+00,  0.00000000E+00,
             2.50000000E-01,  2.50000000E-01,  0.00000000E+00,
             5.00000000E-01,  2.50000000E-01,  0.00000000E+00,
            -2.50000000E-01,  2.50000000E-01,  0.00000000E+00,
             5.00000000E-01,  5.00000000E-01,  0.00000000E+00,
            -2.50000000E-01,  5.00000000E-01,  2.50000000E-01,
            ]
    import numpy as np
    qpoints = np.reshape(qpoints, (-1,3))

    # Global variables used both for the GS and the DFPT run.
    global_vars = dict(
        nband=12,
        ecut=12.0,
        pawecutdg=24.0 if paw else None,
        ngkpt=[8, 8, 8],
        nshiftk=4,
        shiftk=[0.0, 0.0, 0.5,   # This gives the usual fcc Monkhorst-Pack grid
                0.0, 0.5, 0.0,
                0.5, 0.0, 0.0,
                0.5, 0.5, 0.5],
        #shiftk=[0, 0, 0],
        paral_kgb=0,
        diemac=9.0,
    )

    gs_inp = abilab.AbinitInput(structure, pseudos=pseudos)
    gs_inp.set_vars(global_vars, tolvrs=1.0e-18)

    # Get the qpoints in the IBZ. Note that here we use a q-mesh with ngkpt=(4,4,4) and shiftk=(0,0,0)
    # i.e. the same parameters used for the k-mesh in gs_inp.
    #qpoints = gs_inp.abiget_ibz(ngkpt=(4,4,4), shiftk=(0,0,0), kptopt=1).points
    #print("get_ibz", qpoints)

    ph_inp = abilab.AbinitInput(structure, pseudos)

    # Response-function calculation for phonons.
    qpt = [0, 0, 0]
    qpt = [0.25, 0, 0]
    ph_inp.set_vars(
        global_vars,
        rfphon=1,        # Will consider phonon-type perturbation
        nqpt=1,          # One wavevector is to be considered
        qpt=qpt,         # This wavevector is q=0 (Gamma)
        toldfe=1.0e10,
        kptopt=3,
        timopt=-1,
    )
        #rfatpol   1 1   # Only the first atom is displaced
        #rfdir   1 0 0   # Along the first reduced coordinate axis
        #kptopt   2      # Automatic generation of k points, taking

    return gs_inp, ph_inp


def build_flow(options):
    gs_inp, ph_inp = make_inputs()

    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)
    gs_work = flowtk.Work()
    gs_work.register_scf_task(gs_inp)
    flow.register_work(gs_work)
    flow.exclude_from_benchmark(gs_work)

    # Get the list of possible parallel configurations from abinit autoparal.
    max_ncpus, min_eff = options.max_ncpus, options.min_eff
    print("Getting all autoparal confs up to max_ncpus:", max_ncpus, "with efficiency >=", min_eff)

    pconfs = ph_inp.abiget_autoparal_pconfs(max_ncpus, autoparal=1)
    if options.verbose: print(pconfs)

    work = flowtk.Work()
    for conf, omp_threads in product(pconfs, options.omp_list):
        mpi_procs = conf.mpi_ncpus
        if not options.accept_conf(conf, omp_threads): continue

        manager = options.manager.new_with_fixed_mpi_omp(mpi_procs, omp_threads)
        inp = ph_inp.new_with_vars(conf.vars)
        work.register_phonon_task(inp, manager=manager, deps={gs_work[0]: "WFK"})

    print("Found %d configurations" % len(work))
    flow.register_work(work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
AlAs Phonon bands
=================

This example shows how to plot the phonon band structure of AlAs.
See tutorial/lesson_rf2.html
"""
from abipy.abilab import abiopen
import abipy.data as abidata

# Open PHBST file produced by anaddb and extract the phonon bands object.
# (alternatively one can use the shell and `abiopen.py OUT_PHBST.nc -nb`
# to open the file in a jupyter notebook.
with abiopen(abidata.ref_file("trf2_5.out_PHBST.nc")) as ncfile:
    phbands = ncfile.phbands

# Plot the phonon frequencies. Note that the labels for the q-points
# are found automatically in an internal database.
phbands.plot(units="cm-1", title="AlAs without LO-TO splitting")

# Alternatively you can use the optional argument qlabels
# that defines the mapping reduced_coordinates --> name of the q-point.
#qlabels = {
#    (0,0,0): "$\Gamma$",
#    (0.375, 0.375, 0.7500): "K",
#    (0.5, 0.5, 1.0): "X",
#    (0.5, 0.5, 0.5): "L",
#    (0.5, 0.0, 0.5): "X",
#    (0.5, 0.25, 0.75): "W",
#}

# and pass it to the plot method:
#phbands.plot(title="AlAs without LO-TO splitting", qlabels=qlabels)



#!/usr/bin/env python
r"""
Joint Density of States
=======================

This example shows how plot the different contributions
to the electronic joint density of states of Silicon
"""
import abipy.data as abidata
from abipy.abilab import abiopen

# Extract the bands computed with the SCF cycle on a Monkhorst-Pack mesh.
with abiopen(abidata.ref_file("si_scf_WFK.nc")) as wfk_file:
    ebands = wfk_file.ebands

# Select the valence and conduction bands to include in the JDOS
# Here we include valence bands from 0 to 3 and the first conduction band (4).
vrange = range(0, 4)
crange = range(4, 5)

# Plot joint-DOS.
ebands.plot_ejdosvc(vrange, crange)

# Plot decomposition of joint-DOS in terms of v --> c transitions
ebands.plot_ejdosvc(vrange, crange, cumulative=False)

# Show optical (vertical) transitions of energy 2.8 eV
with abiopen(abidata.ref_file("si_nscf_GSR.nc")) as gsr_file:
    gsr_file.ebands.plot_transitions(2.8)



#!/usr/bin/env python
r"""
Interatomic Force Constants
============================

This example shows how to plot the longitudinal part
of the Interatomic Force Constants in real-space
starting from a DDB file.

See e.g. <https://journals.aps.org/prb/pdf/10.1103/PhysRevB.50.13035>

.. important::

    Note that a **manager.yml** configuration file and an abinit installation are required
    to run this script as AbiPy needs to invoke anaddb to compute the IFCs from the DDB file.
"""

#%%
# Open the DDB file for AlAs

import os
import abipy.data as abidata

from abipy import abilab

filepath = os.path.join(abidata.dirpath, "refs", "alas_phonons", "trf2_3_DDB")
ddb = abilab.abiopen(filepath)

#%%
# Call anaddb to compute the Interatomic Force Constants
# Default args are: asr=2, chneut=1, dipdip=1.

ifc = ddb.anaget_ifc()

#%%
# Plot the total longitudinal IFCs in local coordinates,

ifc.plot_longitudinal_ifc(title="Total Longitudinal IFCs")

#%%
# Plot the short range longitudinal IFCs in local coordinates,

ifc.plot_longitudinal_ifc_short_range(title="Longitudinal SR-IFCs")

#%%
# Plot the Ewald part of the IFCs in local coordinates,

ifc.plot_longitudinal_ifc_ewald(title="Longitudinal LR-IFCs")



# coding: utf-8
"""
Shankland-Koelling-Wood Fourier interpolation scheme.
For the theoretical background see :cite:`Euwema1969,Koelling1986,Pickett1988,Madsen2006`.
"""
from __future__ import annotations

import abc
import itertools
import pickle
import numpy as np
import scipy
import time

from collections import deque, OrderedDict
#from typing import
from monty.termcolor import cprint
from monty.collections import dict2namedtuple
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt
from abipy.tools.numtools import gaussian, find_degs_sk
from abipy.core.kpoints import Kpath
from abipy.core.symmetries import mati3inv


class ElectronInterpolator(metaclass=abc.ABCMeta):
    """
    """
    # Tolerances passed to spglib.
    symprec = 1e-5
    angle_tolerance = -1.0

    occtype = "insulator"
    # insulator: fermi level set to homo = nelect//2, occupation factors are either 0 or 1
    #   not suitable for metals, semi-metals, doped semiconductors)
    # fermi-dirac: metallic occupation scheme with physical temperature.
    # gaussian (metallic occupation scheme with gaussian broadening)

    # Disable cache
    use_cache = True

    @classmethod
    def pickle_load(cls, filepath: str):
        """Loads the object from a pickle file."""
        with open(filepath, "rb") as fh:
            return pickle.load(fh)

    def pickle_dump(self, filepath: str) -> None:
        """Save the status of the object in pickle format."""
        with open(filepath, "wb") as fh:
            pickle.dump(self, fh)

    def get_sampling(self, mesh, is_shift):
        """
        Use spglib to compute the k-points in the IBZ with the corresponding weights
        as well as the k-points in the full BZ.

        Args:
            mesh:
            is_shift:

        Return: named tuple with the following attributes:

            ibz:
            nibz
            weights:
            bz:
            nbz
            grid:
        """
        import spglib as spg
        mesh = np.array(mesh)
        mapping, grid = spg.get_ir_reciprocal_mesh(mesh, self.cell,
            is_shift=is_shift, is_time_reversal=self.has_timrev, symprec=self.symprec)

        uniq, weights = np.unique(mapping, return_counts=True)
        weights = np.asarray(weights, dtype=float) / len(grid)
        nkibz = len(uniq)
        ibz = grid[uniq] / mesh
        if self.verbose:
            print("Number of ir-kpoints: %d" % nkibz)

        kshift = 0.0 if is_shift is None else 0.5 * np.asarray(is_shift)
        bz = (grid + kshift) / mesh

        # All k-points and mapping to ir-grid points
        bz2ibz = np.empty(len(bz), dtype=int)
        for i, (ir_gp_id, gp) in enumerate(zip(mapping, grid)):
            inds = np.where(uniq == ir_gp_id)
            #print("inds", inds, "inds[0]", inds[0])
            assert len(inds) == 1
            bz2ibz[i] = int(inds[0])
            #print("%3d ->%3d %s" % (i, ir_gp_id, (gp + [0.5, 0.5, 0.5]) / mesh))
            #print("%3d ->%3d %s" % (i, ir_gp_id, (gp + kshift) / mesh))

        return dict2namedtuple(mesh=mesh, shift=kshift,
                               ibz=ibz, nibz=len(ibz), weights=weights,
                               bz=bz, nbz=len(bz), grid=grid, bz2ibz=bz2ibz)

    #def recalc_fermie(self, kmesh, is_shift=None)
    #    # Compute DOS
    #    edos = _get_cached_edos(kmesh, is_shift)
    #    #if edos is None:
    #    #    edos = self.get_edos(kmesh, is_shift=is_shift, method="gaussian", step=0.1, width=0.2, wmesh=None)
    #    #    self._cache_edos(kmesh, is_shift, edos)
    #    self.fermie = fermie
    #    # Find number of electrons from new chemical potential from nelect.
    #    self.nelect = edos.integral.spline(fermie)
    #    return self.fermie

    @property
    def val_ib(self) -> int:
        """The index of the valence band."""
        if self.occtype != "insulator":
            print("Trying to access valence band index with occtype:", self.occtype)
        return int(self.nelect // 2 - 1)

    #def set_fermie(self, fermie, kmesh, is_shift=None):
    #    """
    #    Change the Fermi level. Use the IDOS computed on the k-grid specifined by
    #    `kmesh` and `is_shift` to recompute the total number of electrons.

    #    Args:
    #        fermie: New Fermi level in eV
    #        kmesh: Three integers with the number of divisions along the reciprocal primitive axes.
    #        is_shift: three integers (spglib API). When is_shift is not None, the kmesh is shifted along
    #            the axis in half of adjacent mesh points irrespective of the mesh numbers. None means unshited mesh.

    #    Return:
    #        New value of `self.nelect`.
    #    """
    #    # Compute DOS
    #    edos = self._get_cached_edos(kmesh, is_shift)
    #    #if edos is None:
    #    #    edos = self.get_edos(kmesh, is_shift=is_shift, method="gaussian", step=0.1, width=0.2, wmesh=None)
    #    #    self._cache_edos(kmesh, is_shift, edos)

    #    self.fermie = fermie
    #    # Find number of electrons from new chemical potential from nelect.
    #    self.nelect = idos.spline(fermie)
    #    return self.nelect

    #def set_nelect(self, nelect, kmesh, is_shift=None):
    #    """
    #    Change the total number of electrons. Use the IDOS computed on the k-grid specifined by
    #    `kmesh` and `is_shift` to recompute the new Fermi level

    #    Args:
    #        nelect: New numbre of electrons.
    #        kmesh: Three integers with the number of divisions along the reciprocal primitive axes.
    #        is_shift: three integers (spglib API). When is_shift is not None, the kmesh is shifted along
    #            the axis in half of adjacent mesh points irrespective of the mesh numbers. None means unshited mesh.

    #    Return:
    #        New value of `self.fermie`.
    #    """
    #    # Compute DOS
    #    edos = self._get_cached_edos(kmesh, is_shift)
    #    #if edos is None:
    #    #    edos = self.get_edos(kmesh, is_shift=is_shift, method="gaussian", step=0.1, width=0.2, wmesh=None)
    #    #    self._cache_edos(kmesh, is_shift, edos)
    #    self.nelect = nelect
    #    # Find new chemical potential from nelect.
    #    idos
    #    self.fermie = new_fermie
    #    return new_fermie

    #def set_charge_per_ucell(self, charge, kmesh, is_shift=None):

    #def get_occfacts(self, eigens, temp):
    #    """
    #    Compute occupation factors from the eigenvalues `eigens` and
    #    the temperature `temp` in K. occfacts in [0, 1].
    #    """
    #    if self.occtype == "insulator":
    #        occfacts = np.ones(eigens.shape)
    #        occfacts[:, :, self.val_ib + 1:] = 0.0
    #    else:
    #        return {
    #            #"gaussian": n_gaussian,
    #            "fermi-dirac": n_fermi_dirac,
    #        }[self.occtype](eigens, self.fermie, temp)

    def get_edos(self, kmesh, is_shift=None, method="gaussian", step=0.1, width=0.2, wmesh=None):
        """
        Compute the electron DOS on a linear mesh.

        Args:
            kmesh: Three integers with the number of divisions along the reciprocal primitive axes.
            is_shift: three integers (spglib API). When is_shift is not None, the kmesh is shifted along
                the axis in half of adjacent mesh points irrespective of the mesh numbers. None means unshited mesh.
            method: String defining the method for the computation of the DOS.
            step: Energy step (eV) of the linear mesh.
            width: Standard deviation (eV) of the gaussian.
            mesh: Frequency mesh to use. If None, the mesh is computed automatically from the eigenvalues.

        Returns:
            (mesh, values, integral)
        """
        k = self.get_sampling(kmesh, is_shift)

        # Interpolate eigenvalues in the IBZ.
        eigens = self._get_cached_eigens(kmesh, is_shift, "ibz")
        if eigens is None:
            eigens = self.interp_kpts(k.ibz).eigens
            self._cache_eigens(kmesh, is_shift, eigens, "ibz")

        # Compute the linear mesh.
        wmesh, step = self._get_wmesh_step(eigens, wmesh, step)
        nw = len(wmesh)
        values = np.zeros((self.nsppol, nw))

        # TODO: Write cython version.
        if method == "gaussian":
            for spin in range(self.nsppol):
                for ik, wtk in enumerate(k.weights):
                    for band in range(self.nband):
                        values[spin] += wtk * gaussian(wmesh, width, center=eigens[spin, ik, band])

            # Compute IDOS
            try :
                from scipy.integrate import cumulative_trapezoid as cumtrapz
            except ImportError:
                from scipy.integrate import cumtrapz

            integral = cumtrapz(values, x=wmesh, initial=0.0)

        else:
            raise ValueError("Method %s is not supported" % method)

        return dict2namedtuple(mesh=wmesh, values=values, integral=integral)
        #return ElectronDos(wmesh, values, integral, is_shift, method, step, width)

    #def get_jdos_q0(self, kmesh, is_shift=None, method="gaussian", step=0.1, width=0.2, wmesh=None):
    #    r"""
    #    Compute the join density of states at q==0

    #        :math:`\sum_{kbv} f_{vk} (1 - f_{ck}) \delta(\omega - E_{ck} + E_{vk})`

    #    Args:
    #        kmesh: Three integers with the number of divisions along the reciprocal primitive axes.
    #        is_shift: three integers (spglib API). When is_shift is not None, the kmesh is shifted along
    #            the axis in half of adjacent mesh points irrespective of the mesh numbers. None means unshited mesh.
    #        method: String defining the method.
    #        step: Energy step (eV) of the linear mesh.
    #        width: Standard deviation (eV) of the gaussian.
    #        wmesh: Frequency mesh to use. If None, the mesh is computed automatically from the eigenvalues.

    #    Returns:
    #    """
    #    k = self.get_sampling(kmesh, is_shift)

    #    # Interpolate eigenvalues in the IBZ.
    #    eigens = self._get_cached_eigens(kmesh, is_shift, "ibz")
    #    if eigens is None:
    #        eigens = self.interp_kpts(k.ibz).eigens
    #        self._cache_eigens(kmesh, is_shift, eigens, "ibz")

    #    wmesh, step = self._get_w2mesh_step(eigens, wmesh, step)
    #    nw = len(wmesh)
    #    values = np.zeros((self.nsppol, nw))

    #    if self.occtype == "insulator":
    #        if method == "gaussian":
    #            for spin in range(self.nsppol):
    #                for ik, wtk in enumerate(k.weights):
    #                    for icb in range(self.val_ib + 1, self.nband):
    #                        ec = eigens[spin, ik, icb]
    #                        for icv in range(self.val_ib):
    #                            ev = eigens[spin, ik, icv]
    #                            values[spin] += wtk * gaussian(wmesh, width, center=ec-ev)
    #        else:
    #            raise ValueError("Method %s is not supported" % method)

    #    else:
    #        raise NotImplementedError("not insulator")
    #        #if method == "gaussian":
    #        #    for spin in range(self.nsppol):
    #        #        for ik, wtk in enumerate(k.weights):
    #        #            for icb in conduction:
    #        #                ec = eigens[spin, ik, icb]
    #        #                fc = 1.0 - occfacts[spin, ik, icb]
    #        #                for icv in valence:
    #        #                    ev = eigens[spin, ik, icv]
    #        #                    fv = occfacts[spin, ik, icv]
    #        #                    fact = wtk * fv * fc
    #        #                    values[spin] += fact * gaussian(wmesh, width, center=ec-ev)
    #        #else:
    #        #    raise ValueError("Method %s is not supported" % method)

    #    if self.nsppol == 1: values *= 2.0
    #    integral = scipy.integrate.cumtrapz(values, x=wmesh, initial=0.0)

    #    return dict2namedtuple(mesh=wmesh, values=values, integral=integral)

    #def get_jdos_qpts(self, qpoints, kmesh, is_shift=None, method="gaussian", step=0.1, width=0.2, wmesh=None):
    #    qpoints = np.reshape(qpoints, (-1, 3))
    #    k = self.get_sampling(kmesh, is_shift)

    #    # Interpolate eigenvalues in the full BZ.
    #    eigens_kbz = self._get_cached_eigens(kmesh, is_shift, "bz")
    #    if eigens is None:
    #        eigens_kbz = self.interp_kpts(k.bz).eigens
    #        self._cache_eigens(kmesh, is_shift, eigens_kbz, "bz")
    #    g_skb = gaussian(eigens_kbz, width)

    #    # TODO: One could reduce the sum to IBZ(q) with appropriate weight.
    #    jdos_sqw = np.zeros((self.nsppol, len(qpoints), nw))
    #    for iq, qpt in enumerate(qpoints):
    #        kpq_bz = k.bz + qpt
    #        eigens_kqbz = self.interp_kpts(kpq_bz).eigens
    #        g_skqb = gaussian(eigens_kqbz, width)
    #        #vals = g_skb * g_skqb
    #        #jdos_sqw[:, iq] = vals.sum(axis=(1, 2))

    #    jdos_sqw *= 1. / k.nbz
    #    return jdos_sqw

    #def get_nesting_at_e0(self, qpoints, kmesh, e0, width=0.2, is_shift=None):
    #    """
    #    Compute the nesting factor with gaussian broadening for an arbitrary list of q-points.

    #    Args:
    #        qpoints: List of q-points in reduced coordinates.
    #        kmesh: Three integers with the number of divisions along the reciprocal primitive axes.
    #        e0: Energy level in eV.
    #        width: Standard deviation (eV) of the gaussian.
    #        is_shift: three integers (spglib API). When is_shift is not None, the kmesh is shifted along
    #            the axis in half of adjacent mesh points irrespective of the mesh numbers. None means unshited mesh.

    #    Returns:
    #        numpy array of shape [self.nsppol, len(qpoints)]
    #    """
    #    qpoints = np.reshape(qpoints, (-1, 3))
    #    k = self.get_sampling(kmesh, is_shift)

    #    # Interpolate eigenvalues in the full BZ.
    #    eigens_kbz = self._get_cached_eigens(kmesh, is_shift, "bz")
    #    if eigens_kbz is None:
    #        eigens_kbz = self.interp_kpts(k.bz).eigens
    #        self._cache_eigens(kmesh, is_shift, eigens_kbz, "bz")

    #    eigens_kbz = eigens_kbz - e0
    #    g_skb = gaussian(eigens_kbz, width)

    #    # TODO: One could reduce the sum to IBZ(q) with appropriate weight.
    #    nest_sq = np.zeros((self.nsppol, len(qpoints)))
    #    for iq, qpt in enumerate(qpoints):
    #        kpq_bz = kbz + qpt
    #        eigens_kqbz = self.interp_kpts(kpq_bz).eigens - e0
    #        g_skqb = gaussian(eigens_kqbz, width)
    #        vals = g_skb * g_skqb
    #        nest_sq[:, iq] = vals.sum(axis=(1, 2))

    #    nest_sq *= 1. / k.nbz
    #    return nest_sq

    def _get_wmesh_step(self, eigens, wmesh, step):
        if wmesh is not None:
            return wmesh, wmesh[1] - wmesh[0]

        # Compute the linear mesh.
        epad = 1.0
        e_min = eigens.min() - epad
        e_max = eigens.max() + epad
        nw = int(1 + (e_max - e_min) / step)
        wmesh, step = np.linspace(e_min, e_max, num=nw, endpoint=True, retstep=True)

        return wmesh, step

    def _get_w2mesh_step(self, eigens, wmesh, step):
        if wmesh is not None:
            return wmesh, wmesh[1] - wmesh[0]

        # Compute the linear mesh.
        cmin, cmax = +np.inf, -np.inf
        vmin, vmax = +np.inf, -np.inf
        for spin in range(self.nsppol):
            for c in range(self.val_ib + 1, self.nband):
                cmin = min(cmin, eigens[spin,:,c].min())
                cmax = max(cmax, eigens[spin,:,c].max())
            for v in range(self.val_ib + 1):
                vmin = min(vmin, eigens[spin,:,v].min())
                vmax = max(vmax, eigens[spin,:,v].max())
        #cmin = eigens.max()
        #cmax = cmin
        #vmax = eigens.min()
        #vmin = vmax

        e_min = cmin - vmax
        e_min -= 0.1 * abs(e_min)
        e_max = cmax - vmin
        e_max += 0.1 * abs(e_max)

        nw = int(1 + (e_max - e_min) / step)
        wmesh, step = np.linspace(e_min, e_max, num=nw, endpoint=True, retstep=True)

        return wmesh, step

    def _get_cached_eigens(self, kmesh, is_shift, kzone):
        """
        Return a copy of the interpolated eigenvalues associated to (kmesh, is_shift, kzone).
        Return None if eigens are not available.
        """
        if not self.use_cache: return None
        if not hasattr(self, "_cached_eigens"): self._cached_eigens = OrderedDict()
        kmesh = tuple(kmesh)
        if is_shift is not None: is_shift = tuple(is_shift)
        arr = self._cached_eigens.get((kmesh, is_shift, kzone))
        if arr is not None: arr = arr.copy()
        return arr

    def _cache_eigens(self, kmesh, is_shift, eigens, kzone):
        """
        Save interpolated eigenvalues associated to (kmesh, is_shift, kzone).
        """
        if not self.use_cache: return None
        if not hasattr(self, "_cached_eigens"): self._cached_eigens = OrderedDict()
        kmesh = tuple(kmesh)
        if is_shift is not None: is_shift = tuple(is_shift)
        self._cached_eigens[(kmesh, is_shift, kzone)] = eigens.copy()

    def _get_cached_edos(self, kmesh, is_shift):
        """
        Return the electron DOS obtained from the interpolated eigenvalues associated
        to (kmesh, is_shift). None if DOS is not available.
        """
        if not self.use_cache: return None
        if not hasattr(self, "_cached_edos"): self._cached_edos = OrderedDict()
        kmesh = tuple(kmesh)
        if is_shift is not None: is_shift = tuple(is_shift)
        return self._cached_edos.get((kmesh, is_shift))

    def _cache_edos(self, kmesh, is_shift, edos):
        """
        Save the electron DOS obtained from the interpolated eigenvalues associated to (kmesh, is_shift).
        """
        if not self.use_cache: return None
        if not hasattr(self, "_cached_edos"): self._cached_edos = OrderedDict()
        kmesh = tuple(kmesh)
        if is_shift is not None: is_shift = tuple(is_shift)
        self._cached_edos[(kmesh, is_shift)] = edos # .copy()

    @add_fig_kwargs
    def plot_dos_vs_kmeshes(self, kmeshes, is_shift=None, method="gaussian", step=0.1, width=0.2,
                            fontsize=12, ax=None, **kwargs):
        """
        Plot (interpolated) DOSes computed with different meshes.

        Args:
            kmeshes: List of kmeshes. Each item is given by three integers with the number of
                divisions along the reciprocal primitive axes.
            is_shift: three integers (spglib_ API). When is_shift is not None, the kmesh is shifted along
                the axis in half of adjacent mesh points irrespective of the mesh numbers. None means unshited mesh.
            method: String defining the method for the computation of the DOS.
            step: Energy step (eV) of the linear mesh.
            width: Standard deviation (eV) of the gaussian.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            fontsize: Legend and title fontsize.

        Returns: |matplotlib-Figure|
        """
        ax, fig, plt = get_ax_fig_plt(ax=ax)

        kmeshes = np.reshape(np.asarray(kmeshes, dtype=int), (-1, 3))
        for kmesh in kmeshes:
            edos = self.get_edos(kmesh, is_shift=is_shift, method=method, step=step, width=width)
            for spin in range(self.nsppol):
                spin_sign = +1 if spin == 0 else -1
                ax.plot(edos.mesh, edos.values[spin] * spin_sign, label=str(kmesh) if spin == 0 else None)

        ax.grid(True)
        ax.set_xlabel("Energy (eV)")
        ax.set_ylabel('DOS (states/eV)')
        ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    #@add_fig_kwargs
    #def plot_jdosq0_vs_kmeshes(self, kmeshes, is_shift=None, method="gaussian", step=0.1, width=0.2,
    #                           ax=None, fontsize=12, **kwargs):
    #    """
    #    Plot (interpolated) Joint DOSes at q=0 computed with different meshes.

    #    Args:
    #        kmeshes: List of kmeshes. Each item is given by three integers with the number of
    #            divisions along the reciprocal primitive axes.
    #        is_shift: three integers (spglib API). If None, the kmesh is shifted along
    #            the axis in half of adjacent mesh points irrespective of the mesh numbers. None means unshited mesh.
    #        method: String defining the method for the computation of the DOS.
    #        step: Energy step (eV) of the linear mesh.
    #        width: Standard deviation (eV) of the gaussian.
    #        ax: |matplotlib-Axes| or None if a new figure should be created.
    #        fontsize: Legend and title fontsize.

    #    Returns: |matplotlib-Figure|
    #    """
    #    ax, fig, plt = get_ax_fig_plt(ax=ax)
    #    kmeshes = np.reshape(np.asarray(kmeshes, dtype=int), (-1, 3))
    #    for kmesh in kmeshes:
    #        jdos = self.get_jdos_q0(kmesh, is_shift=is_shift, method=method, step=step, width=width)
    #        for spin in range(self.nsppol):
    #            spin_sign = +1 if spin == 0 else -1
    #            ax.plot(jdos.mesh, jdos.values[spin] * spin_sign, label=str(kmesh) if spin == 0 else None)

    #    ax.grid(True)
    #    ax.set_xlabel("Energy (eV)")
    #    ax.set_ylabel('JDOS (states/eV)')
    #    ax.legend(loc="best", fontsize=fontsize, shadow=True)

    #    return fig

    #@add_fig_kwargs
    #def plot_nesting_vs_widths(self, widths, kmesh, e0=None, qvertices_names=None,
    #                           line_density=20, is_shift=None, ax=None, fontsize=12, **kwargs):
    #    """
    #    Plot (interpolated) nesting factor computed with different broadening.

    #    Args:
    #        widths: List of standard deviations (eV) of the gaussian.
    #        kmeshes: List of kmeshes. Each item is given by three integers with the number of
    #            divisions along the reciprocal primitive axes.
    #        e0: Energy level in eV.
    #        qvertices_names
    #        line_density:
    #        is_shift: three integers (spglib API). When is_shift is not None, the kmesh is shifted along
    #            the axis in half of adjacent mesh points irrespective of the mesh numbers. None means unshited mesh.
    #        ax: |matplotlib-Axes| or None if a new figure should be created.
    #        fontsize: Legend and title fontsize.

    #    Returns: |matplotlib-Figure|
    #    """
    #    ax, fig, plt = get_ax_fig_plt(ax=ax)
    #    qpoints = self._get_kpts_kticks_klabels(ax, qvertices_names, line_density)

    #    e0 = self.interpolated_fermie if e0 is None else e0
    #    for width in np.asarray(widths):
    #        nest_sq = self.get_nesting_at_e0(qpoints, kmesh, e0, width=width, is_shift=is_shift)
    #        for spin in range(self.nsppol):
    #            spin_sign = +1 if spin == 0 else -1
    #            ax.plot(nest_sq[spin] * spin_sign, label=str(kmesh) if spin == 0 else None)

    #    ax.grid(True)
    #    ax.set_ylabel('Nesting factor')
    #    ax.legend(loc="best", fontsize=fontsize, shadow=True)

    #    return fig

    #@add_fig_kwargs
    #def plot_nesting_vs_kmeshes(self, width, kmeshes, e0=None, qvertices_names=None, line_density=20,
    #                            is_shift=None, ax=None, fontsize=12, **kwargs):
    #    """
    #    Plot (interpolated) nesting factor computed with different k-meshes.

    #    Args:
    #        width: Gaussian broadening (eV).
    #        kmeshes: List of kmeshes. Each item is given by three integers with the number of
    #            divisions along the reciprocal primitive axes.
    #        e0: Energy level in eV.
    #        qvertices_names
    #        line_density:
    #        is_shift: three integers (spglib API). When is_shift is not None, the kmesh is shifted along
    #            the axis in half of adjacent mesh points irrespective of the mesh numbers. None means unshited mesh.
    #        ax: |matplotlib-Axes| or None if a new figure should be created.
    #        fontsize: legend and title fontsize.

    #    Returns: |matplotlib-Figure|
    #    """
    #    ax, fig, plt = get_ax_fig_plt(ax=ax)
    #    qpoints = self._get_kpts_kticks_klabels(ax, qvertices_names, line_density)

    #    kmeshes = np.reshape(np.asarray(kmeshes, dtype=int), (-1, 3))
    #    e0 = self.interpolated_fermie if e0 is None else e0
    #    for kmesh in kmeshes:
    #        nest_sq = self.get_nesting_at_e0(qpoints, kmesh, e0, width=width, is_shift=is_shift)
    #        for spin in range(self.nsppol):
    #            spin_sign = +1 if spin == 0 else -1
    #            ax.plot(nest_sq[spin] * spin_sign, label=str(kmesh) if spin == 0 else None)

    #    ax.grid(True)
    #    ax.set_ylabel('Nesting factor')
    #    ax.legend(loc="best", fontsize=fontsize, shadow=True)

    #    return fig

    #@add_fig_kwargs
    #def plot_group_velocites(self, vertices_names=None, line_density=20, ax=None, **kwargs):
    #    """
    #    Plot (interpolated) group velocities computed along an arbitrary k-path.

    #    Args:
    #        vertices_names
    #        line_density:
    #        ax: |matplotlib-Axes| or None if a new figure should be created.

    #    Returns: |matplotlib-Figure|
    #    """
    #    ax, fig, plt = get_ax_fig_plt(ax=ax)
    #    kfrac_coords = self._get_kpts_kticks_klabels(ax, vertices_names, line_density)

    #    #v_skb = self.interp_kpts(kfrac_coords, dk1=v_skb)
    #    for spin in range(self.nsppol):
    #        for band in range(self.nband):
    #            # plot |v|
    #            v_skb[spin, :, band]
    #            ax.plot(vals, color="k" if spin == 0 else "r")

    #    ax.grid(True)
    #    ax.set_ylabel('Group Velocities')

    #    return fig

    def _get_kpts_kticks_klabels(self, ax, vertices_names, line_density):
        if vertices_names is None:
            vertices_names = [(k.frac_coords, k.name) for k in self.structure.hsym_kpoints]

        kpath = Kpath.from_vertices_and_names(self.structure, vertices_names, line_density=line_density)

        #self._axset_ticks_labels(ax, xticks, xlabels)
        return kpath.frac_coords, list(range(len(kpath))), kpath.names

    #@abc.abstractproperty
    #def structure(self):
    #    """Returns the |Structure| object."""

    #nsppol
    #nkpt
    #nband
    #original_fermie
    #fermie
    #verbose
    #has_timrev

    @abc.abstractmethod
    def eval_sk(self, spin, kpt, der1=None, der2=None):
        """
        Interpolate eigenvalues for all bands at a given (spin, k-point).
        Optionally compute gradients and Hessian matrices.

        Args:
            spin: Spin index.
            kpt: K-point in reduced coordinates.
            der1: If not None, ouput gradient is stored in der1[nband, 3].
            der2: If not None, output Hessian is der2[nband, 3, 3].

        Return:
            oeigs[nband]
        """

    def interp_kpts(self, kfrac_coords, dk1=False, dk2=False):
        """
        Interpolate energies on an arbitrary set of k-points. Optionally, compute
        gradients and Hessian matrices.

        Args:
            kfrac_coords: K-points in reduced coordinates.
            dk1 (bool): True if gradient is wanted.
            dk2 (bool): True to compute 2nd order derivatives.

        Return:
            namedtuple with:
            interpolated energies in eigens[nsppol, len(kfrac_coords), nband]
            gradient in dedk[self.nsppol, len(kfrac_coords), self.nband, 3))
            hessian in dedk2[self.nsppol, len(kfrac_coords), self.nband, 3, 3))

            gradient and hessian are set to None if not computed.
        """
        start = time.time()

        kfrac_coords = np.reshape(kfrac_coords, (-1, 3))
        new_nkpt = len(kfrac_coords)
        new_eigens = np.empty((self.nsppol, new_nkpt, self.nband))

        dedk = None if not dk1 else np.empty((self.nsppol, new_nkpt, self.nband, 3))
        dedk2 = None if not dk2 else np.empty((self.nsppol, new_nkpt, self.nband, 3, 3))

        der1, der2 = None, None
        for spin in range(self.nsppol):
            for ik, newk in enumerate(kfrac_coords):
                if dk1: der1 = dedk[spin, ik]
                if dk2: der2 = dedk2[spin, ik]
                new_eigens[spin, ik] = self.eval_sk(spin, newk, der1=der1, der2=der2)

        if self.verbose:
            print("Interpolation completed in %.3f (s)" % (time.time() - start))

        return dict2namedtuple(eigens=new_eigens, dedk=dedk, dedk2=dedk2)

    def interp_kpts_and_enforce_degs(self, kfrac_coords, ref_eigens, atol=1e-4):
        """
        Interpolate energies on an arbitrary set of k-points. Use `ref_eigens`
        to detect degeneracies and average the interpolated values in the degenerate subspace.
        """
        kfrac_coords = np.reshape(kfrac_coords, (-1, 3))
        new_nkpt = len(kfrac_coords)
        ref_eigens = np.reshape(ref_eigens, (self.nsppol, new_nkpt, self.nband))

        # Interpolate eigenvales.
        new_eigens = self.interp_kpts(kfrac_coords).eigens

        # Average interpolated values over degenerates bands.
        for spin in range(self.nsppol):
            for ik in range(new_nkpt):
                for dgbs in find_degs_sk(ref_eigens[spin, ik], atol):
                    if len(dgbs) == 1: continue
                    new_eigens[spin, ik, dgbs] = new_eigens[spin, ik, dgbs].sum() / len(dgbs)

        return dict2namedtuple(eigens=new_eigens, dedk=None, dedk2=None)

    #def get_ebands3d(self, mesh, is_shift):
    #    k = self.get_sampling(self, mesh, is_shift)
    #    ibz_eigens = self.interp_kpts(k.ibz).eigens
    #    ibz_kpoints
    #    return ElectronBands3D(self.structure, ibz_kpoints, self.has_timrev, ibz_eigens, self.fermie)


class SkwInterpolator(ElectronInterpolator):
    """
    This object implements the Shankland-Koelling-Wood Fourier interpolation scheme.
    It can be used to interpolate functions in k-space with the periodicity of the
    reciprocal lattice and satisfying F(k) = F(Sk) for each rotation S
    belonging to the point group of the crystal. For readability reason,
    the names of the variables are chosen assuming we are interpolating electronic eigenvalues
    but the same object can be used to interpolate other quantities. Just set the first dimension to 1.
    """

    def __init__(self, lpratio, kpts, eigens, fermie, nelect, cell, symrel, has_timrev,
                 filter_params=None, verbose=1):
        """
        Args:
            lpratio: Ratio between the number of star-functions and the number of ab-initio k-points.
                5-10 should be OK in many systems, larger values may be required for accurate derivatives.
            kpts: numpy array with the [nkpt, 3] ab-initio k-points in reduced coordinates.
            eigens: numpy array with the ab-initio energies. shape [nsppol, nkpt, nband].
            fermie: Fermi energy in eV.
            nelect: Number of electrons in the unit cell
            cell: (lattice, positions, numbers)
                lattice: numpy array with direct lattice vectors along the rows.
                positions: atomic positions in reduced coordinates.
                numbers: Atomic number for each atom in positions.
            symrel: [nsym, 3, 3] numpy array with the (ferromagnetic) symmetry operations of the direct lattice
                in reduced coordinates. anti-ferromagnetic symmetries (if any) should be removed by the caller.
            has_timrev: True is time-reversal can be used.
            filter_params: List with parameters used to filter high-frequency components (Eq 9 of PhysRevB.61.1639)
                First item gives rcut, second item sigma. Ignored if None.
            verbose: Verbosity level.
        """
        self.verbose = verbose
        self.cell = cell
        lattice = self.cell[0]
        self.original_fermie = fermie
        self.interpolated_fermie = self.original_fermie
        self.nelect = nelect
        self.has_timrev = has_timrev

        # iscomplexobj is used to handle lifetimes.
        eigens = np.atleast_3d(eigens)
        self.iscomplexobj = np.iscomplexobj(eigens)
        self.nsppol, self.nkpt, self.nband = eigens.shape

        if len(kpts) != self.nkpt:
            raise ValueError("Second dimension of eigens should be %d but got array of shape: %s" %
                (len(kpts), eigens.shape))
        if self.nkpt == 1:
            raise ValueError("Interpolation algorithm requires nkpt > 1")

        rprimd = np.asarray(lattice).T
        self.rmet = np.matmul(rprimd.T, rprimd)

        # Find point group operations.
        symrel = np.reshape(symrel, (-1, 3, 3))
        self.ptg_symrel, self.ptg_symrec, has_inversion = extract_point_group(symrel, has_timrev)
        self.ptg_nsym = len(self.ptg_symrel)
        if self.verbose:
            print("Found", self.ptg_nsym, "symmetries in point group")

        # Find nrwant star points.
        self.lpratio = lpratio = int(lpratio)
        if lpratio <= 1:
            raise ValueError("lpratio must be > 1 but got %s" % lpratio)

        nrwant = lpratio * self.nkpt
        fact = 1/2 if has_inversion else 1
        rmax = int((1.0 + (lpratio * self.nkpt * self.ptg_nsym * fact) / 2.0) ** (1/3.)) * np.ones(3, dtype=int)
        #rmax = int((1.0 + (lpratio * self.nkpt) / 2.0) ** (1/3.)) * np.ones(3, dtype=int)

        while True:
            self.rpts, r2vals, ok = self._find_rstar_gen(nrwant, rmax)
            self.nr = len(self.rpts)
            if ok:
                break
            else:
                print("rmax: ", rmax," was not large enough to find", nrwant, "R-star points.")
                rmax *= 2
                print("Will try again with enlarged rmax:", rmax)

        print("Using:", self.nr, "star-functions. nstars/nk:", self.nr / self.nkpt)

        # Compute (inverse) roughness function.
        c1, c2 = 0.25, 0.25
        r2min = r2vals[1]
        inv_rhor = 1.0 / ((1.0 - c1 * r2vals / r2min) ** 2 + c2 * (r2vals / r2min) ** 3)

        # Construct star functions for the ab-initio k-points.
        nsppol, nband, nkpt, nr = self.nsppol, self.nband, self.nkpt, self.nr
        self.skr = np.empty((nkpt, nr), dtype=complex)
        for ik, kpt in enumerate(kpts):
            self.skr[ik] = self.get_stark(kpt)

        # Build H(k,k') matrix (Hermitian)
        hmat = np.empty((nkpt-1, nkpt-1), dtype=complex)
        for jk in range(nkpt-1):
            v_jkr = self.skr[jk, 1:] - self.skr[nkpt-1, 1:]
            #for ik in range(jk + 1):
            for ik in range(nkpt-1):
                v_ikr = inv_rhor[1:] * (self.skr[ik, 1:] - self.skr[nkpt-1, 1:])
                hmat[ik, jk] = np.vdot(v_jkr, v_ikr)
                if ik == jk: hmat[ik, jk] = hmat[ik, jk].real

        # Solving system of linear equations to get lambda coeffients (eq. 10 of PRB 38 2721)..."
        de_kbs = np.empty((nkpt-1, nband, nsppol), dtype=complex)
        for spin in range(nsppol):
            for ib in range(nband):
                de_kbs[:, ib, spin] = eigens[spin, 0:nkpt-1, ib] - eigens[spin, nkpt-1, ib]

        # Solve all bands and spins at once
        # FIXME: Portability problem with scipy 0.19 in which linalg.solve wraps the expert drivers
        # http://scipy.github.io/devdocs/release.0.19.0.html#foreign-function-interface-improvements
        if scipy.__version__ == "0.19.0":
            import warnings
            warnings.warn("linalg.solve in scipy 0.19.0 gives weird results. Use at your own risk!!!")

        try:
            lmb_kbs = scipy.linalg.solve(hmat, np.reshape(de_kbs, (-1, nband * nsppol)))
            #lmb_kbs = scipy.linalg.solve(hmat, np.reshape(de_kbs, (-1, nband * nsppol)),
                    #sym_pos=True, lower=False, overwrite_a=True, overwrite_b=True, check_finite=False)
                    #sym_pos=False, lower=False, overwrite_a=False, overwrite_b=False, check_finite=True)

        except scipy.linalg.LinAlgError as exc:
            print("Cannot solve system of linear equations to get lambda coeffients (eq. 10 of PRB 38 2721)")
            print("This usually happens when there are symmetrical k-points passed to the interpolator.")
            raise exc

        lmb_kbs = np.reshape(lmb_kbs, (-1, nband, nsppol))

        # Compute coefficients.
        self.coefs = np.empty((nsppol, nband, nr), dtype=complex)
        for spin in range(nsppol):
            for ib in range(nband):
                for ir in range(1, nr):
                    self.coefs[spin, ib, ir] = inv_rhor[ir] \
                        * np.vdot(self.skr[:nkpt-1, ir] - self.skr[nkpt-1, ir], lmb_kbs[:nkpt-1, ib, spin])

                self.coefs[spin, ib, 0] = eigens[spin, nkpt-1, ib] \
                    - np.dot(self.coefs[spin, ib, 1:nr], self.skr[nkpt-1, 1:nr])

        # Filter high-frequency.
        self.rcut, self.rsigma = None, None
        if filter_params is not None:
            self.rcut = filter_params[0] * np.sqrt(r2vals[-1])
            self.rsigma = rsigma = filter_params[1]
            if self.verbose:
                print("Applying filter (Eq 9 of PhysRevB.61.1639) with rcut:", self.rcut, ", rsigma", self.rsigma)
            from scipy.special import erfc
            for ir in range(1, nr):
                self.coefs[:, :, ir] *= 0.5 * erfc((np.sqrt(r2vals[ir]) - self.rcut) / self.rsigma)

        # Prepare workspace arrays for star functions.
        self.cached_kpt = np.ones(3) * np.inf
        self.cached_kpt_dk1 = np.ones(3) * np.inf
        self.cached_kpt_dk2 = np.ones(3) * np.inf

        # Compare ab-initio data with interpolated results.
        mae = 0.0
        for spin in range(nsppol):
            for ik, kpt in enumerate(kpts):
                skw_eb = self.eval_sk(spin, kpt)
                mae += np.abs(eigens[spin, ik] - skw_eb).sum()
                if self.verbose >= 10:
                    # print interpolated eigenvales
                    for band in range(self.nband):
                        e0 = eigens[spin, ik, band]
                        eskw = skw_eb[band]
                        print("spin", spin, "band", band, "ikpt", ik, "e0", e0, "eskw", eskw, "diff", e0 - eskw)

        mae *= 1e3 / (nsppol * nkpt * nband)
        if np.isnan(mae) or np.isinf(mae) or mae > 1000:
            raise RuntimeError("Interpolation went bananas! mae = %s" % mae)

        warn = mae > 10.0
        cprint("FIT vs input data: Mean Absolute Error= %.3e (meV)" % mae, color="red" if warn else "green")
        if warn:
            # Issue warning if error too large.
            cprint("Large error in SKW interpolation!", color="red")
            cprint(f"MAE: {mae} [meV]", color="red")

        self.mae = mae

    def __str__(self):
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation."""
        lines = []
        app = lines.append
        app("nsppol: %s, nband: %s" % (self.nsppol, self.nband))
        app("Number of operations in point-group: %s with time-reversal: %s" % (self.ptg_nsym, self.has_timrev))
        app("Number of ab-initio k-points: %s" % self.nkpt)
        app("Number of star functions: %s [nstars/nk = %s]" % (self.nr, (self.nr / self.nkpt)))
        if self.rcut is not None:
            app("Fourier filter (Eq 9 of PhysRevB.61.1639) with rcut: %s, rsigma: %s" % (self.rcut, self.rsigma))
        app("Comparison between ab-initio data and fit gave Mean Absolute Error: %s [meV]" % self.mae)

        return "\n".join(lines)

    def eval_sk(self, spin, kpt, der1=None, der2=None) -> np.ndarray:
        """
        Interpolate eigenvalues for all bands at a given (spin, k-point).
        Optionally compute gradients and Hessian matrices.

        Args:
            spin: Spin index.
            kpt: K-point in reduced coordinates.
            der1: If not None, ouput gradient is stored in der1[nband, 3].
            der2: If not None, output Hessian is der2[nband, 3, 3].

        Return:
            oeigs[nband]
        """
        # [NB, NR] x [NR]
        oeigs = np.matmul(self.coefs[spin], self.get_stark(kpt))
        if not self.iscomplexobj: oeigs = oeigs.real

        if der1 is not None:
            skr_dk1 = self.get_stark_dk1(kpt)
            for ii in range(3):
                value = np.matmul(self.coefs[spin, :, :], skr_dk1[ii])
                if not self.iscomplexobj: value = value.real
                der1[:, ii] = value

        if der2 is not None:
            skr_dk2 = self.get_stark_dk2(kpt)
            for jj in range(3):
                for ii in range(jj + 1):
                    value = np.matmul(self.coefs[spin, :, :], skr_dk2[ii,jj])
                    if not self.iscomplexobj: value = value.real
                    der2[:, ii, jj] = value
                    if ii != jj: der2[jj, ii] = der2[ii, jj]

        return oeigs

    #def eval_skb(self, spin, kpt, band, der1=None, der2=None):
    #    """
    #    Interpolate eigenvalues for a given (spin, k-point, band).

    #    Args:

    #    Return:
    #    """
    #    # Compute star function for this k-point (if not already in memory)
    #    if np.any(kpt != self.cached_kpt):
    #        self.cached_skr, self.cached_kpt = self.get_stark(kpt), kpt

    #    # Do not take complex conjugate.
    #    oeig = np.dot(self.coefs[spin, band], self.cached_skr)
    #    if not self.iscomplexobj: oeigs = oeigs.real
    #    if der1 is None and der2 is None: return oeig

    #    # TODO: Test Derivatives
    #    if der1 is not None:
    #        # Compute first-order derivatives.
    #        if np.any(kpt != self.cached_kpt_dk1):
    #            self.cached_skr_dk1, self.cached_kpt_dk1 = self.get_stark_dk1(kpt), kpt

    #        for ii in range(3):
    #            value = np.dot(self.coefs[spin, band], self.cached_skr_dk1[ii])
    #            if not self.iscomplexobj: value = value.real
    #            der1[ii] = value

    #        if der2 is None: return oeig, der1

    #    if der2 is not None:
    #        # Compute second-order derivatives.
    #        if np.any(kpt != self.cached_kpt_dk2):
    #            self.cached_skr_dk2, self.cached_kpt_dk2 = self.get_stark_dk2(kpt), kpt

    #        der2 = zero
    #        for jj in range(3):
    #            for ii in range(jj + 1):
    #                value = np.dot(self.coefs[spin, band], self.cached_skr_dk2[ii,jj])
    #                if not self.iscomplexobj: value = value.real
    #                der2[ii, jj] = value
    #                if ii != jj: der2[jj, ii] = der2[ii, jj]

    #    return oeig, der1, der2

    def get_stark(self, kpt) -> np.ndarray:
        """
        Return the star function for k-point `kpt`.

        Args:
            kpt: K-point in reduced coordinates.

        Return:
            complex array of shape [self.nr]
        """
        two_pi = 2.0 * np.pi
        skr = np.zeros(self.nr, dtype=complex)
        _np_exp = np.exp
        for omat in self.ptg_symrel:
            sk = two_pi * np.matmul(omat.T, kpt)
            skr += _np_exp(1.j * np.matmul(self.rpts, sk))
        skr /= self.ptg_nsym

        return skr

    def get_stark_dk1(self, kpt) -> np.ndarray:
        """
        Compute the 1st-order derivative of the star function wrt k

        Args:
            kpt: K-point in reduced coordinates.

        Return:
            complex array [3, self.nr]  with the derivative of the
            star function wrt k in reduced coordinates.
        """
        srk_dk1 = np.zeros((3, self.nr), dtype=complex)
        two_pi = 2.0 * np.pi
        rpts_t = self.rpts.T

        for omat in self.ptg_symrel:
            sk = two_pi * np.matmul(omat.T, kpt)
            exp_skr = np.exp(1.j * np.matmul(self.rpts, sk))
            for ir, rr in enumerate(self.rpts):
                srk_dk1[:, ir] += exp_skr[ir] * np.matmul(omat, rr)
            #for ir, or in enumerate(np.matmul(omat, rpts_t).T):
            #    srk_dk1[:, ir] += exp_skr[ir] * or

        srk_dk1 *= 1.j / self.ptg_nsym
        return srk_dk1

    def get_stark_dk2(self, kpt) -> np.ndarray:
        """
        Compute the 2nd-order derivatives of the star function wrt k.

        Args:
            kpt: K-point in reduced coordinates.

        Return:
            Complex numpy array of shape [3, 3, self.nr] with the 2nd-order derivatives
            of the star function wrt k in reduced coordinates.
        """
        srk_dk2 = np.zeros((3, 3, self.nr), dtype=complex)
        raise NotImplementedError()
        #work = zero
        #do isym=1,self.ptg_nsym
        #   sk = two_pi * matmul(transpose(self%ptg_symrel(:,:,isym)), kpt)
        #   do ir=1,self%nr
        #     sr = matmul(self%ptg_symrel(:,:,isym), self%rpts(:,ir))
        #     eiskr = exp(j_dpc * dot_product(sk, self%rpts(:,ir)))
        #     do jj=1,3
        #        do ii=1,jj
        #            work(ii,jj,ir) = work(ii,jj,ir) + eiskr * sr(ii) * sr(jj)

        #work = - work / self.ptg_nsym

        #do jj=1,3
        #   do ii=1,jj
        #       srk_dk2(:, ii, jj) = work(ii, jj, :)
        #       if (ii /= jj) srk_dk2(:,jj,ii) = work(:,ii,jj)

        return srk_dk2

    #def find_stationary_points(self, kmesh, bstart=None, bstop=None, is_shift=None)
    #    k = self.get_sampling(kmesh, is_shift)
    #    if bstart is None: bstart = self.nelect // 2 - 1
    #    if bstop is None: bstop = self.nelect // 2
    #    nb = bstop - bstart + 1
    #    results = []
    #    for ik_ibz, kpt in enumerate(k.ibz):
    #        vk_b = self.eval_dk1(kpt, bstart, bstop)
    #        bands = []
    #        for ib, v in enumerate(vk_b):
    #            if v < atol: bands.append(ib + bstart)
    #        if bands:
    #            #results.append()
    #            for band in bands:
    #                d2k_b = self.eval_dk2(kpt, band)

    #    return results

    def _find_rstar_gen(self, nrwant, rmax) -> tuple:
        """
        Find all lattice points generating the stars inside the supercell defined by `rmax`

        Args:
            nrwant: Number of star-functions required.
            rmax: numpy array with the maximum number of cells along the 3 reduced directions.

        Returns:
            tuple: (rpts, r2vals, ok)
        """
        msize = (2 * rmax + 1).prod()
        rtmp = np.empty((msize, 3), dtype=int)
        r2tmp = np.empty(msize)
        if self.verbose: print("rmax", rmax, "msize:", msize)

        start = time.time()
        for cnt, l in enumerate(itertools.product(range(-rmax[0], rmax[0] + 1),
                                                  range(-rmax[1], rmax[1] + 1),
                                                  range(-rmax[2], rmax[2] + 1))):
            rtmp[cnt] = l
            r2tmp[cnt] = np.dot(l, np.matmul(self.rmet, l))

        if self.verbose: print("gen points", time.time() - start)

        start = time.time()
        # Sort r2tmp and rtmp
        iperm = np.argsort(r2tmp)
        r2tmp = r2tmp[iperm]
        rtmp = rtmp[iperm]
        #return rtmp, r2tmp, True

        # Find shells
        r2sh = np.empty(msize, dtype=int)    # Correspondence between R and shell index.
        shlim = np.empty(msize, dtype=int)   # For each shell, the index of the initial G-vector.
        nsh = 1
        r2sh[0] = 0
        shlim[0] = 0
        r2_prev = 0.0
        for ir in range(1, msize):
            if abs(r2tmp[ir] - r2_prev) > r2tmp[ir] * 1e-8:
                r2_prev = r2tmp[ir]
                shlim[nsh] = ir
                nsh += 1
            r2sh[ir] = nsh
        shlim[nsh] = msize
        if self.verbose:
            print("nshells", nsh)
            print("shells", time.time() - start)

        # Find R-points generating the stars.
        # This is the most CPU critical part. I think it's difficult to do better than this without cython
        start = time.time()
        rgen = deque()

        for ish in range(nsh):
            ss, ee = shlim[ish], shlim[ish + 1]
            if ss + 1 == ee:
                rgen.append(rtmp[ss])
                continue

            if True:
                # This is faster
                rs = set([tuple(rtmp[ss])])
                for ir in range(ss + 1, ee):
                    rvec = rtmp[ir]
                    if all(tuple(np.matmul(rot, rvec)) not in rs for rot in self.ptg_symrel):
                        rs.add(tuple(rvec))
            else:
                rs = deque([rtmp[ss]])
                for ir in range(ss + 1, ee):
                    for rot in self.ptg_symrel:
                        rot_r = np.matmul(rot, rtmp[ir])
                        if any(np.all(rot_r == x) for x in rs): break
                    else:
                        # We have new R-point.
                        #print("Adding new point")
                        rs.append(rtmp[ir])

            #print(len(rs), rs)
            rgen.extend(rs)
        #print(rgen)
        if self.verbose: print("stars", time.time() - start)

        start = time.time()
        rgen = np.array(rgen, dtype=int)
        nstars = len(rgen)

        # Store rpts and compute ||R||**2.
        ok = nstars >= nrwant
        nr = min(nstars, nrwant)
        rpts = rgen[:nr].copy()
        r2vals = np.empty(nr)
        for ir in range(nr):
            r2vals[ir] = np.dot(rpts[ir], np.matmul(self.rmet, rpts[ir]))

        if self.verbose:
            print("r2max ", rpts[nr-1])
            print("end ", time.time() - start)
            if self.verbose > 10:
                print("nstars:", nstars)
                for r, r2 in zip(rpts, r2vals):
                    print(r, r2)

        return rpts, r2vals, ok


def extract_point_group(symrel, has_timrev) -> tuple:
    """
    Extract the point group rotations from the spacegroup. Add time-reversal
    if spatial inversion is not present and `has_timrev`.

    Return:
        (ptg_symrel, ptg_symrec) with rotations in real- and reciprocal-space.
    """
    nsym = len(symrel)
    tmp_nsym = 1
    work_symrel = np.empty((2*nsym, 3, 3), dtype=int)
    work_symrel[0] = symrel[0]

    for isym in range(1, nsym):
        found = any(np.all(symrel[isym] == work_symrel[search]) for search in range(tmp_nsym))
        if not found:
            work_symrel[tmp_nsym] = symrel[isym]
            tmp_nsym += 1

    inversion_3d = -np.eye(3, dtype=int)
    has_inversion = any(np.all(w == inversion_3d) for w in work_symrel[:tmp_nsym])

    # Now we know the symmetries of the point group.
    ptg_nsym = 2 * tmp_nsym if not has_inversion and has_timrev else tmp_nsym
    ptg_symrel = np.empty((ptg_nsym, 3, 3), dtype=int)
    ptg_symrec = np.empty((ptg_nsym, 3, 3), dtype=int)

    ptg_symrel[:tmp_nsym] = work_symrel[:tmp_nsym]
    for isym in range(tmp_nsym):
        ptg_symrec[isym] = mati3inv(ptg_symrel[isym])

    if not has_inversion and has_timrev:
        # Add inversion.
        ptg_symrel[tmp_nsym:] = -work_symrel[:tmp_nsym]
        for isym in range(tmp_nsym, ptg_nsym):
            ptg_symrec[isym] = mati3inv(ptg_symrel[isym])

    return ptg_symrel, ptg_symrec, has_inversion



#!/usr/bin/env python
r"""
Flow for quasi-harmonic calculations
====================================

Warning: This code is still under development.
"""
import sys
import os
import abipy.abilab as abilab
import abipy.data as abidata

from abipy import flowtk
from abipy.flowtk.qha import QhaFlow


def build_flow(options):
    """
    Create a `QhaFlow` for quasi-harmonic calculations.
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        __file__ = os.path.join(os.getcwd(), "run_qha.py")
        options.workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    # Initialize structure and pseudos
    structure = abilab.Structure.from_file(abidata.cif_file("si.cif"))
    pseudos = abidata.pseudos("14si.pspnc")

    # Build input for GS calculation.
    scf_input = abilab.AbinitInput(structure, pseudos)
    scf_input.set_vars(ecut=12, nband=8, tolvrs=1e-8)
    scf_input.set_kmesh(ngkpt=[4, 4, 4], shiftk=[0, 0, 0])

    v0 = scf_input.structure.volume
    volumes = [0.08 * v0, v0, v0 * 1.02]
    return QhaFlow.from_scf_input(options.workdir, scf_input, volumes,
                                  ngqpt=[2, 2, 2], with_becs=False, edos_ngkpt=(4, 4, 4))


# This block generates the thumbnails in the Abipy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Raman with BSE and frozen phonon
================================

This script shows how to perform a Raman calculation with
excitonic effects included with the BSE formalism.
"""

import sys
import os
import numpy as np
import abipy.abilab as abilab
import abipy.data as abidata
from abipy import flowtk


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_","flow_")

    flow = flowtk.Flow(options.workdir, manager=options.manager)

    pseudos = abidata.pseudos("14si.pspnc")

    # Get the unperturbed structure.
    base_structure = abidata.structure_from_ucell("Si")

    etas = [-.001, 0, +.001]
    ph_displ = np.reshape(np.zeros(3*len(base_structure)), (-1,3))
    ph_displ[0,:] = [+1, 0, 0]
    ph_displ[1,:] = [-1, 0, 0]

    # Build new structures by displacing atoms according to the phonon displacement
    # ph_displ (in cartesian coordinates). The Displacement is normalized so that
    # the maximum atomic diplacement is 1 Angstrom and then multiplied by eta.
    modifier = abilab.StructureModifier(base_structure)

    displaced_structures = modifier.displace(ph_displ, etas, frac_coords=False)

    # Generate the different shifts to average
    ndiv = 2
    shift1D = np.arange(1,2*ndiv+1,2)/(2*ndiv)
    all_shifts = [[x,y,z] for x in shift1D for y in shift1D for z in shift1D]
    all_shifts = [[0, 0, 0]]

    for structure, eta in zip(displaced_structures, etas):
        for shift in all_shifts:
            flow.register_work(raman_work(structure, pseudos, shift))

    return flow


def raman_work(structure, pseudos, shiftk, paral_kgb=1):
    # Generate 3 different input files for computing optical properties with BSE.

    # Global variables
    global_vars = dict(
        ecut=8,
        istwfk="*1",
        chksymbreak=0,
        #nstep=4,
        nstep=10,
        paral_kgb=paral_kgb,
    )

    # GS run
    scf_inp = abilab.AbinitInput(structure, pseudos=pseudos)
    scf_inp.set_vars(global_vars)
    scf_inp.set_kmesh(ngkpt=[2, 2, 2], shiftk=shiftk)
    scf_inp["tolvrs"] = 1e-6

    # NSCF run
    nscf_inp = abilab.AbinitInput(structure, pseudos=pseudos)
    nscf_inp.set_vars(global_vars)
    nscf_inp.set_kmesh(ngkpt=[2, 2, 2], shiftk=shiftk)

    nscf_inp.set_vars(
        tolwfr=1e-8,
        nband=12,
        nbdbuf=4,
        iscf=-2,
    )

    # BSE run with Model dielectric function and Haydock (only resonant + W + v)
    # Note that SCR file is not needed here
    bse_inp = abilab.AbinitInput(structure, pseudos=pseudos)
    bse_inp.set_vars(global_vars)
    bse_inp.set_kmesh(ngkpt=[2, 2, 2], shiftk=shiftk)

    bse_inp.set_vars(
        optdriver=99,
        ecutwfn=global_vars["ecut"],
        ecuteps=3,
        inclvkb=2,
        bs_algorithm=2,       # Haydock
        bs_haydock_niter=4,   # No. of iterations for Haydock
        bs_exchange_term=1,
        bs_coulomb_term=21,   # Use model W and full W_GG.
        mdf_epsinf=12.0,
        bs_calctype=1,        # Use KS energies and orbitals to construct L0
        mbpt_sciss="0.8 eV",
        bs_coupling=0,
        bs_loband=2,
        nband=6,
        #bs_freq_mesh="0 10 0.1 eV",
        bs_hayd_term=0,      # No terminator
    )

    # Build the work representing a BSE run with model dielectric function.
    return flowtk.BseMdfWork(scf_inp, nscf_inp, bse_inp)


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_raman_bse.py -s
#
# then use:
#
#    abirun.py flow_raman_bse robot MDF
#
# to analyze all the macroscopic dielectric functions produced by the Flow with the robot.
# Inside the ipython terminal type:
#
# .. code-block:: ipython
#
#   [1] %matplotlib
#   [2] robot.plot()
#
# to compare the real and imaginary part of the macroscopic dielectric function for the different displacements.
#
# .. image:: https://github.com/abinit/abipy_assets/blob/master/run_raman_bse.png?raw=true
#    :alt: Band structure of Si in the IBZ and along a k-path
#



#!/usr/bin/env python
r"""
spin-polarized fatbands
=======================

This example shows how to plot the L-projected fatbands of Ni
using the results stored in the FATBANDS.nc files produced with prtdos 3.
"""

#%%
# Open the file (alternatively one can use the shell and `abiopen.py FILE -nb`
# to open the file in a jupyter notebook
# This file has been produced on a k-path so it's not suitable for DOS calculations.

import abipy.abilab as abilab
import abipy.data as abidata

fbnc_kpath = abilab.abiopen(abidata.ref_file("ni_kpath_FATBANDS.nc"))

#%%
# NC files have contributions up to L = 4 (g channel)
# but here we are intererested in s,p,d terms only so we use the optional argument lmax

lmax = 2

# Energy limits in eV for plots. The pseudo contains semi-core states but
# we are not interested in this energy region. Fermi level set to zero.
elims = [-10, 2]

# Print file info (dimensions, variables ...)
# Note that prtdos = 3, so LM decomposition is not available.
print(fbnc_kpath)

# Plot the k-points belonging to the path.
#fbnc_kpath.ebands.kpoints.plot()

#%%
# Plot the electronic fatbands grouped by atomic type.

fbnc_kpath.plot_fatbands_typeview(ylims=elims, lmax=lmax, tight_layout=True)

#%%
# For the plotly version use:

fbnc_kpath.plotly_fatbands_typeview(ylims=elims, lmax=lmax)

#%%
# Plot the electronic fatbands grouped by L.

fbnc_kpath.plot_fatbands_lview(ylims=elims, lmax=lmax, tight_layout=True)

#%%
# For the plotly version use:

fbnc_kpath.plotly_fatbands_lview(ylims=elims, lmax=lmax)

#%%
# Now we read another FATBANDS file produced on 18x18x18 k-mesh

fbnc_kmesh = abilab.abiopen(abidata.ref_file("ni_666k_FATBANDS.nc"))

#%%
# Plot the L-PJDOS grouped by atomic type.

fbnc_kmesh.plot_pjdos_typeview(xlims=elims, lmax=lmax, tight_layout=True)

#%%
# For the plotly version use:

fbnc_kmesh.plotly_pjdos_typeview(xlims=elims, lmax=lmax)

#%%
# Plot the L-PJDOS grouped by L.

fbnc_kmesh.plot_pjdos_lview(xlims=elims, lmax=lmax, tight_layout=True)

#%%
# For the plotly version use:

fbnc_kmesh.plotly_pjdos_lview(xlims=elims, lmax=lmax)

#%%
# Now we use the two netcdf files to produce plots with fatbands + PJDOSEs.
# The data for the DOS is taken from pjdosfile.

fbnc_kpath.plot_fatbands_with_pjdos(pjdosfile=fbnc_kmesh,
                                    ylims=elims, lmax=lmax, view="type", tight_layout=True)

#%%
# For the plotly version use:

fbnc_kpath.plotly_fatbands_with_pjdos(pjdosfile=fbnc_kmesh,
                                      ylims=elims, lmax=lmax, view="type")

#%%
# fatbands + PJDOS grouped by L

fbnc_kpath.plot_fatbands_with_pjdos(pjdosfile=fbnc_kmesh,
                                    ylims=elims, lmax=lmax, view="lview", tight_layout=True)

#%%
# For the plotly version use:

fbnc_kpath.plotly_fatbands_with_pjdos(pjdosfile=fbnc_kmesh,
                                      ylims=elims, lmax=lmax, view="lview")

#%%
# Close files.

fbnc_kpath.close()
fbnc_kmesh.close()



#!/usr/bin/env python
"""
This script runs all the python scripts located in this directory
"""
from __future__ import annotations

import sys
import os
import argparse
import shutil
import tempfile
import abipy.tools.cli_parsers as cli

from subprocess import call
from abipy import __version__
from abipy import flowtk



def main():
    def str_examples():
        examples = """
          Usage example:\n\n
          _runflows.py => Run all scripts.
        """
        return examples

    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(str_examples())
        if err_msg:
            sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = argparse.ArgumentParser(epilog=str_examples(),formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-V', '--version', action='version', version="%(prog)s version " + __version__)
    parser.add_argument('--loglevel', default="ERROR", type=str,
                        help="set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")

    parser.add_argument('-m', '--mode', type=str, default="sequential", help="execution mode. Default is sequential.")
    parser.add_argument('-e', '--exclude', type=str, default="", help="Exclude scripts. Comma-separated names")
    parser.add_argument('-x', '--execute', default=False, action="store_true", help="Execute flows.")
    parser.add_argument('--keep-dirs', action="store_true", default=False, help="Do not remove flow directories.")
    parser.add_argument('-b', '--bail-on-failure', default=False, help="Exit at the first error.")

    options = parser.parse_args()

    cli.set_loglevel(options.loglevel)

    # Find scripts.
    if options.exclude:
        options.exclude = options.exclude.split(",")
        print("Will exclude:\n", options.exclude)

    root = os.path.abspath(os.path.join(os.path.dirname(__file__), "flows"))
    scripts = []
    for fname in os.listdir(root):
        if fname in options.exclude: continue
        if fname.endswith(".py") and fname.startswith("run_"):
            path = os.path.join(root, fname)
            if path != __file__:
                scripts.append(path)
    print("Executing: %d scripts with mode: `%s` and execute: `%s`" % (
          len(scripts), options.mode, options.execute))

    # Run scripts according to mode.
    dirpaths, errors, retcode, cnt = [], [], 0, 0
    if options.mode in ["s", "sequential"]:
        for script in scripts:
            # flow will be produced in a temporary workdir.
            workdir = tempfile.mkdtemp(prefix='flow_' + os.path.basename(script))
            ret = call(["python", script, "--workdir", workdir])
            retcode += ret

            if ret != 0:
                e = "python %s returned retcode != 0" % script
                print(e)
                errors.append(e)
                if options.bail_on_failure:
                    print("Exiting now since bail_on_failure")
                    break

            dirpaths.append(workdir)

            # Here we execute the flow
            if options.execute:
                cnt += 1
                ret = 0
                try:
                    flow = flowtk.Flow.pickle_load(workdir)
                    flow.make_scheduler().start()
                    if not flow.all_ok: retcode += 1

                except Exception as exc:
                    ret += 1
                    s = "Exception raised during flow execution: %s\n:%s" % (flow, exc)
                    print(s)
                    errors.append(s)
                    if options.bail_on_failure:
                        print("Exiting now since bail_on_failure")
                        break
                    retcode += ret

        # Remove directories.
        if not options.keep_dirs:
            for dirpath in dirpaths:
                try:
                    shutil.rmtree(dirpath, ignore_errors=False)
                except OSError:
                    print("Exception while removing %s" % dirpath)

    else:
        show_examples_and_exit(err_msg="Wrong value for mode: %s" % options.mode)

    if errors:
        for i, err in enumerate(errors):
            print(92 * "=")
            print("[%d] %s" % (i, err))
            print(92 * "=")

    print("Number of scripts executed %d/%d" % (cnt, len(scripts)), ", final retcode %d" % retcode)
    return retcode


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
"""Work subclasses for the computation of luminiscent properties."""
from __future__ import annotations

from .works import Work
from abipy.abilab import abiopen
from abipy.lumi.deltaSCF import DeltaSCF


class LumiWork(Work):
    """
    This Work implements Fig 1 of https://arxiv.org/abs/2010.00423.

    Client code is responsible for the preparation of the supercell and
    of the GS SCF input files with the fixed electronic occupations associated to the two configurations.
    By default, the work computes the two relaxed structures and the four total energies
    corresponding to the Ag, Ag*, Ae*, Ae configurations. Optionally, one can activate the computation of
    four electronic band structures. See docstring of from_scf_inputs for further info.
    """

    @classmethod
    def from_scf_inputs(cls, gs_scf_inp, ex_scf_inp, relax_kwargs_gs, relax_kwargs_ex, ndivsm=0, nb_extra=10,
                        tolwfr=1e-12, four_points=True, meta=None, manager=None) -> LumiWork:
        """
        Args:
            gs_scf_inp: |AbinitInput| representing a GS SCF run for the ground-state.
            ex_scf_inp: |AbinitInput| representing a GS SCF run for the excited-state.
            relax_kwargs_gs: Dictonary with input variables to be added to gs_scf_inp
                when generating input files for ground state structural relaxations.
            relax_kwargs_ex: Dictonary with input variables to be added to ex_scf_inp
                when generating input files for excited state structural relaxations.
            ndivsm: Activates the computation of band structure if different from zero.
                if > 0, it's the number of divisions for the smallest segment of the path (Abinit variable).
                if < 0, it's interpreted as the pymatgen `line_density` parameter in which the number of points
                in the segment is proportional to its length. Typical value: -20.
                This option is the recommended one if the k-path contains two high symmetry k-points that are very close
                as ndivsm > 0 may produce a very large number of wavevectors.
            nb_extra: Number of extra bands added to the input nband when computing band structures (ndivsm != 0).
            tolwfr: Tolerance of the residuals used for the NSCF band structure calculations.
            four_points : if True, compute the two relaxations and the four points energies.
                If false, only the two relaxations.
            meta : dict corresponding to the metadata of a lumiwork (supercell size, dopant type,...)
            manager: |TaskManager| of the task. If None, the manager is initialized from the config file.
        """
        new = cls(manager=manager)

        # Templates for GS SCF calculations.
        new.gs_scf_inp = gs_scf_inp
        new.ex_scf_inp = ex_scf_inp

        # Save paramenters for the generation of input files at runtime.
        new.relax_kwargs_gs = relax_kwargs_gs
        new.relax_kwargs_ex = relax_kwargs_ex

        new.four_points = four_points
        new.ndivsm = int(ndivsm)
        new.tolwfr = tolwfr
        new.nb_extra = int(nb_extra)
        new.meta=meta

        # Relaxation for the Ag configuration.`
        new.gs_relax_task = new.register_relax_task(gs_scf_inp.new_with_vars(relax_kwargs_gs))

        # Internal counter used in on_all_ok to drive the differ steps of the calculations.
        new.iteration_step = 0

        # JSON-compatible dictionary storing the most important results of the Work. (relaxations and 4 pts results
        # are separated)
        # Results will be stored in the `lumi_4pts.json` file in the outdata directory of the Work
        # so that one can easily implement additional post-processing tools.
        new.json_data = {}

        return new

    def on_all_ok(self):
        """
        This method is called when all the works in the flow have reached S_OK.

        This is the section in which we implement most of the workflow logic at runtime.
        since we need to generate input files with relaxed structures.
        """
            # Get Ag relaxed structure
        #with self.gs_relax_task.open_gsr() as gsr:
           # ag_relaxed_structure = gsr.structure
        with abiopen(self.gs_relax_task.output_file.path) as relax_gs_abo:
            ag_relaxed_structure = relax_gs_abo.final_structure

        if self.iteration_step == 0:
            print("in iteration step 0")
            self.iteration_step += 1

            # Relax geometry with excited configuration starting from Ag*.
            relax_ex_inp=self.ex_scf_inp.new_with_vars(self.relax_kwargs_ex)
            relax_ex_inp_2 = relax_ex_inp.new_with_structure(ag_relaxed_structure)
            self.ex_relax_task = self.register_relax_task(relax_ex_inp_2)#,deps={self.gs_relax_task: "DEN"})

            # if only the two relaxation, go to results writing step directly
            if self.four_points==False:
                self.iteration_step = 2

            return self.postpone_on_all_ok()

        elif self.iteration_step == 1:
            print("in iteration step 1")
            self.iteration_step += 1

            ##### SCF task for each configuration and optionnaly their electronic band structures ####

            # Build GS SCF input for the Ag configuration:
            # use same structure as Ag with ground occupation factors.
            ag_scf_inp = self.gs_scf_inp.new_with_structure(ag_relaxed_structure)
            self.ag_scf_task = self.register_scf_task((ag_scf_inp),deps={self.gs_relax_task: "DEN"})

            # Build GS SCF input for the Ag* configuration:
            # use same structure as Ag but with excited occupation factors.
            agstar_scf_inp = self.ex_scf_inp.new_with_structure(ag_relaxed_structure)
            self.agstar_scf_task = self.register_scf_task((agstar_scf_inp),deps={self.gs_relax_task: "DEN"})

            # Get Aestar relaxed structure.
            #with self.ex_relax_task.open_gsr() as gsr:
            #    aestar_relaxed_structure = gsr.structure
            with abiopen(self.ex_relax_task.output_file.path) as relax_ex_abo:
                aestar_relaxed_structure = relax_ex_abo.final_structure

            # Build ex SCF input for the Aestar configuration:
            # use same structure as Aestar with excited occupation factors.
            aestar_scf_inp = self.ex_scf_inp.new_with_structure(aestar_relaxed_structure)
            self.aestar_scf_task = self.register_scf_task((aestar_scf_inp),deps={self.ex_relax_task: "DEN"})

            # Build GS SCF task for the Ae configuration:
            # use same structure as Aestar but with ground occupation factors.
            ae_scf_inp = self.gs_scf_inp.new_with_structure(aestar_relaxed_structure)
            self.ae_scf_task = self.register_scf_task((ae_scf_inp),deps={self.ex_relax_task: "DEN"})

            if self.ndivsm != 0:
                # Compute band structure for Ag configuration.
                self.ag_scf_task.add_ebands_task_to_work(self, ndivsm=self.ndivsm,
                                                             tolwfr=self.tolwfr, nb_extra=self.nb_extra)

                # Compute band structure for Agstar configuration.
                self.agstar_scf_task.add_ebands_task_to_work(self, ndivsm=self.ndivsm,
                                                             tolwfr=self.tolwfr, nb_extra=self.nb_extra)

                # Compute band structure for aestar configuration.
                self.aestar_scf_task.add_ebands_task_to_work(self, ndivsm=self.ndivsm,
                                                             tolwfr=self.tolwfr, nb_extra=self.nb_extra)

                # Compute band structure for Ae configuration.
                self.ae_scf_task.add_ebands_task_to_work(self, ndivsm=self.ndivsm,
                                                             tolwfr=self.tolwfr, nb_extra=self.nb_extra)

            return self.postpone_on_all_ok()

        elif self.iteration_step == 2:

            print("in iteration step 2")
            self.iteration_step += 1
            ##### Writing the results in json files #####

            self.json_data["meta"] = self.meta

            #with self.gs_relax_task.open_gsr() as gsr:
            self.json_data["gs_relax_filepath"]=self.gs_relax_task.gsr_path

            #with self.ex_relax_task.open_gsr() as gsr:
            self.json_data["ex_relax_filepath"]=self.ex_relax_task.gsr_path

            if self.four_points == True:
                # Get Ag total energy.
                #with self.ag_scf_task.open_gsr() as gsr:
                self.json_data["Ag_gsr_filepath"] = self.ag_scf_task.gsr_path

                # Get Agstar total energy.
                #with self.ex_relax_task.open_gsr() as gsr:
                self.json_data["Agstar_gsr_filepath"] = self.agstar_scf_task.gsr_path

                # Get Aestar total energy.
                #with self.aestar_scf_task.open_gsr() as gsr:
                self.json_data["Aestar_gsr_filepath"] = self.aestar_scf_task.gsr_path

                # Get Aestar total energy.
                #with self.ae_scf_task.open_gsr() as gsr:
                self.json_data["Ae_gsr_filepath"] = self.ae_scf_task.gsr_path

            # Write json file in the outdir of the work
            self.write_json_in_outdir("lumi.json", self.json_data)

            # Build deltascf results 
            delta_scf = DeltaSCF.from_four_points_file([self.ag_scf_task.gsr_path,
                                                        self.agstar_scf_task.gsr_path,
                                                        self.aestar_scf_task.gsr_path,
                                                        self.ae_scf_task.gsr_path])

            # Create dict with all post-processed results
            d = delta_scf.get_dict_results()

            # save d in json format.
            self.write_json_in_outdir("Delta_SCF.json", d)

            return super().on_all_ok()


class LumiWork_relaxations(Work):
    """
    This Work implements the ground and excited state relaxations only.

    The relaxations run simultaneously. No task creation at run-time. 
    """

    @classmethod
    def from_scf_inputs(cls, gs_scf_inp, ex_scf_inp, relax_kwargs_gs, relax_kwargs_ex, meta=None, manager=None):
        """
        Args:
            gs_scf_inp: |AbinitInput| representing a GS SCF run for the ground-state.
            ex_scf_inp: |AbinitInput| representing a GS SCF run for the excited-state.
            relax_kwargs_gs: Dictonary with input variables to be added to gs_scf_inp
                when generating input files for ground state structural relaxations.
            relax_kwargs_ex: Dictonary with input variables to be added to ex_scf_inp
                when generating input files for excited state structural relaxations.
            meta : dict corresponding to the metadata of a lumiwork (supercell size, dopant type,...)
            manager: |TaskManager| of the task. If None, the manager is initialized from the config file.
        """
        new = cls(manager=manager)

        # Templates for GS SCF calculations.
        new.gs_scf_inp = gs_scf_inp
        new.ex_scf_inp = ex_scf_inp

        # Save paramenters for the generation of input files at runtime.
        new.relax_kwargs_gs = relax_kwargs_gs
        new.relax_kwargs_ex = relax_kwargs_ex

        new.meta=meta

        # Relaxation for the Ag configuration.
        new.gs_relax_task = new.register_relax_task(gs_scf_inp.new_with_vars(relax_kwargs_gs))
        new.ex_relax_task = new.register_relax_task(ex_scf_inp.new_with_vars(relax_kwargs_ex))

        new.json_data = {}

        return new

    def on_all_ok(self):
        """
        This method is called when all the works in the flow have reached S_OK.

        """
        self.json_data["meta"] = self.meta

        # with self.gs_relax_task.open_gsr() as gsr:
        self.json_data["gs_relax_filepath"] = self.gs_relax_task.gsr_path

        # with self.ex_relax_task.open_gsr() as gsr:
        self.json_data["ex_relax_filepath"] = self.ex_relax_task.gsr_path

        # Write json file in the outdir of the work
        self.write_json_in_outdir("lumi_relaxations.json", self.json_data)

        return super().on_all_ok()


class LumiWorkFromRelax(Work):
    """
    Same as LumiWork, without the relaxations. Typically used after a LumiWork_relaxations work.
    The two relaxed structures (in ground and excited state) are given as input. No creation at run-time

    """
    @classmethod
    def from_scf_inputs(cls, gs_scf_inp, ex_scf_inp, gs_structure, ex_structure, ndivsm=0, nb_extra=10,
                        tolwfr=1e-12, meta=None ,manager=None):
        """
        Args:
            gs_scf_inp: |AbinitInput| representing a GS SCF run for the ground-state.
            exc_scf_inp: |AbinitInput| representing a GS SCF run for the excited-state.
            gs_structure: object representing the relaxed ground state structure
            ex_structure: object representing the excited ground state structure
            ndivsm: Activates the computation of band structure if different from zero.
                if > 0, it's the number of divisions for the smallest segment of the path (Abinit variable).
                if < 0, it's interpreted as the pymatgen `line_density` parameter in which the number of points
                in the segment is proportional to its length. Typical value: -20.
                This option is the recommended one if the k-path contains two high symmetry k-points that are very close
                as ndivsm > 0 may produce a very large number of wavevectors.
            nb_extra: Number of extra bands added to the input nband when computing band structures (ndivsm != 0).
            tolwfr: Tolerance of the residuals used for the NSCF band structure calculations.
            manager: |TaskManager| of the task. If None, the manager is initialized from the config file.
        """
        new = cls(manager=manager)

        new.meta=meta
        new.gs_structure=gs_structure
        new.ex_structure=ex_structure

        # Templates for GS SCF calculations.
        new.gs_scf_inp = gs_scf_inp
        new.ex_scf_inp = ex_scf_inp

        new.ndivsm = int(ndivsm)
        new.tolwfr = tolwfr
        new.nb_extra = int(nb_extra)

        # JSON-compatible dictionary storing the most important results of the Work. (relaxations and 4 pts results
        # are separated)
        # Results will be stored in the `lumi_4pts.json` or 'lumi_relax.json file in the outdata directory of the Work
        # so that one can easily implement additional post-processing tools.
        new.json_data = {}

        # Build GS SCF input for the Ag configuration:
        # use same structure as Ag but with ground occupation factors
        ag_scf_inp = new.gs_scf_inp.new_with_structure(gs_structure)
        new.ag_scf_task = new.register_scf_task(ag_scf_inp)

        # Build GS SCF input for the Ag* configuration:
        # use same structure as Ag but with excited occupation factors.
        agstar_scf_inp = new.ex_scf_inp.new_with_structure(gs_structure)
        new.agstar_scf_task = new.register_scf_task(agstar_scf_inp)

        # Build ex SCF input for the Aestar configuration:
        aestar_scf_inp = new.ex_scf_inp.new_with_structure(ex_structure)
        new.aestar_scf_task = new.register_scf_task(aestar_scf_inp)

        # Build GS SCF input for the Ag* configuration:
        # use same structure as Ag but with excited occupation factors.
        ae_scf_inp = new.gs_scf_inp.new_with_structure(ex_structure)
        new.ae_scf_task = new.register_scf_task(ae_scf_inp)

        if new.ndivsm != 0:
            # Compute band structure for Ag configuration.
            new.ag_scf_task.add_ebands_task_to_work(new, ndivsm=new.ndivsm,
                                                         tolwfr=new.tolwfr, nb_extra=new.nb_extra)
            # Compute band structure for Ag* configuration.
            new.agstar_scf_task.add_ebands_task_to_work(new, ndivsm=new.ndivsm,
                                                             tolwfr=new.tolwfr, nb_extra=new.nb_extra)
            # Compute band structure for aestar configuration.
            new.aestar_scf_task.add_ebands_task_to_work(new, ndivsm=new.ndivsm,
                                                         tolwfr=new.tolwfr, nb_extra=new.nb_extra)
            # Compute band structure for Ae configuration.
            new.ae_scf_task.add_ebands_task_to_work(new, ndivsm=new.ndivsm,
                                                             tolwfr=new.tolwfr, nb_extra=new.nb_extra)

        return new

    def on_all_ok(self):

        self.json_data["meta"] = self.meta

        # Get Ag total energy.
        # with self.ag_scf_task.open_gsr() as gsr:
        self.json_data["Ag_gsr_filepath"] = self.ag_scf_task.gsr_path

        # Get Agstar total energy.
        # with self.ex_relax_task.open_gsr() as gsr:
        self.json_data["Agstar_gsr_filepath"] = self.agstar_scf_task.gsr_path

        # Get Aestar total energy.
        # with self.aestar_scf_task.open_gsr() as gsr:
        self.json_data["Aestar_gsr_filepath"] = self.aestar_scf_task.gsr_path

        # Get Aestar total energy.
        # with self.ae_scf_task.open_gsr() as gsr:
        self.json_data["Ae_gsr_filepath"] = self.ae_scf_task.gsr_path

        # Write json file in the outdir of the work
        self.write_json_in_outdir("lumi.json", self.json_data)

        # Build deltascf results 
        delta_scf = DeltaSCF.from_four_points_file([self.ag_scf_task.gsr_path,
                                                        self.agstar_scf_task.gsr_path,
                                                        self.aestar_scf_task.gsr_path,
                                                        self.ae_scf_task.gsr_path])

        # Create dict with all post-processed results
        d = delta_scf.get_dict_results()
        # save d in json format.
        self.write_json_in_outdir("Delta_SCF.json", d)

        return super().on_all_ok()


"""
This module defines enumerators associated to important Abinit input variables
"""
from __future__ import annotations

import enum

class EnumMixin:
    """Mixin for enums provides extra capabilities."""

    @classmethod
    def validate(cls, value) -> None:
        """Validate value"""
        values = [member.value for member in cls]
        if value not in values:
            raise ValueError(f"{value=} is not valid. Must be among: {values=}")


class StrEnum(str, enum.Enum):  # StrEnum added in 3.11
    def __new__(cls, *args):
        for arg in args:
            if not isinstance(arg, (str, enum.auto)):
                raise TypeError(
                    "Values of StrEnums must be strings: {} is a {}".format(
                        repr(arg), type(arg)
                    )
                )
        return super().__new__(cls, *args)

    def __str__(self):
        return self.value

    # The first argument to this function is documented to be the name of the
    # enum member, not `self`:
    # https://docs.python.org/3.6/library/enum.html#using-automatic-values
    def _generate_next_value_(name, *_):
        return name


class RUNL(EnumMixin, enum.IntEnum):
    """
    Values of optdriver corresponding to the different run-levels. See defs_basis.F90
    """
    GSTATE     = 0
    RESPFN     = 1
    SCREENING  = 3
    SIGMA      = 4
    NONLINEAR  = 5
    GWR        = 6
    EPH        = 7
    WFK        = 8
    RTTDDFT    = 9
    GWLS       = 66
    BSE        = 99
    LONGWAVE   = 10

    def __str__(self):
        return str(self.value)


class WFK_TASK(EnumMixin, enum.IntEnum):
    """
    Integer flags defining the task to be performed in wfk_analyze. See defs_basis.F90
    """
    NONE      = 0
    FULLBZ    = 1
    CLASSIFY  = 2
    PAW_AEPSI = 3
    EINTERP   = 4
    DDK       = 5
    DDK_DIAGO = 6
    OPTICS_FULLBZ = 7
    KPTS_ERANGE= 8
    CHECK_SYMTAB = 9

    def __str__(self):
        return str(self.value)


class GWR_TASK(EnumMixin, StrEnum):  # StrEnum added in 3.11
    """
    String flags defining the task to be performed in the GWR code.
    """
    HDIAGO = "HDIAGO"
    HDIAGO_FULL = "HDIAGO_FULL"
    CC4S = "CC4S"
    CC4S_FULL = "CC4S_FULL"
    G0W0 = "G0W0"
    G0V = "G0V"
    EGEW = "EGEW"
    EGW0 = "EGW0"
    G0EW = "G0EW"
    RPA_ENERGY = "RPA_ENERGY"
    GAMMA_GW = "GAMMA_GW"
    CHI0 = "CHI0"



"""Panels for interacting with FATBANDS.nc files."""
from __future__ import annotations

#import param
import panel as pn
import panel.widgets as pnw

from abipy.electrons.fatbands import FatBandsFile
from .core import PanelWithElectronBands, ply, mpl, dfc, depends_on_btn_click #, PanelWithEbandsRobot


class FatBandsFilePanel(PanelWithElectronBands):
    """
    Panel with widgets to interact with a |FatBandsFile|.
    """

    def __init__(self, ncfile: FatBandsFile, **params):
        PanelWithElectronBands.__init__(self, ebands=ncfile.ebands, **params)
        self.ncfile = ncfile

        # Create buttons
        self.plot_fatbands_btn = pnw.Button(name="Plot fatbands", button_type='primary')
        self.plot_fatdos_btn = pnw.Button(name="Plot fatdos", button_type='primary')

    @depends_on_btn_click('plot_fatbands_btn')
    def on_plot_fatbands_btn(self) -> pn.Column:
        """
        Plot fatbands grouped by atomic type and angular momentum l
        """
        sz_mode = "stretch_width"
        col = pn.Column(sizing_mode=sz_mode); ca = col.append

        # Plot the electronic fatbands grouped by atomic type.
        ca("## Electronic fatbands grouped by atomic type")
        fig = self.ncfile.plotly_fatbands_typeview(e0="fermie", fact=1.0, lmax=None, fig=None, ylims=None,
                                                   blist=None, fontsize=12, band_and_dos=0, show=False)
        ca(ply(fig))

        # Plot the electronic fatbands grouped by l
        ca("## Electronic fatbands grouped by angular momentum l")
        fig = self.ncfile.plotly_fatbands_lview(e0="fermie", fact=1.0, lmax=None, fig=None, ylims=None,
                                                blist=None, fontsize=12, band_and_dos=0, show=False)
        ca(ply(fig))

        return col

    @depends_on_btn_click('plot_fatdos_btn')
    def on_plot_fatdos_btn(self) -> pn.Column:
        """
        Plot PJDOS grouped by atomic type and angular momentum l
        """
        sz_mode = "stretch_width"
        col = pn.Column(sizing_mode=sz_mode); ca = col.append

        # Plot the L-PJDOS grouped by atomic type.
        lmax = 2
        ca("## Electronic fatdos grouped by atomic type:")
        fig = self.ncfile.plotly_pjdos_typeview(lmax=lmax, show=False)
        ca(ply(fig))

        # Plot the L-PJDOS grouped by L.
        ca("## Electronic fatdos grouped by L:")
        fig = self.ncfile.plotly_pjdos_lview(lmax=lmax, show=False)
        ca(ply(fig))

        return col

    def get_panel(self, as_dict=False, **kwargs):
        """Return tabs with widgets to interact with the FATBANDS.nc file."""
        d = {}

        d["Summary"] = self.get_summary_view_for_abiobj(self.ncfile)
        d["e-Bands"] = self.get_plot_ebands_view()

        if self.ncfile.ebands.kpoints.is_ibz:
            # Add DOS tab but only if we have a k-sampling.
            d["e-DOS"] = self.get_plot_edos_view()

            d["FatDos"] = pn.Row(
                self.pws_col(["## Fatdos", "plot_fatdos_btn"]),
                self.on_plot_fatdos_btn
            )

            if not self.ebands.isnot_ibz_sampling():
                d["ifermi"] = self.get_ifermi_view()

        elif self.ebands.kpoints.is_path:
            # NC files have contributions up to L=4 (g channel)
            # but here we are intererested in s,p,d terms only so
            # we use the optional argument lmax
            lmax = 2

            d["FatBands"] = pn.Row(
                self.pws_col(["## Fatbands", "plot_fatbands_btn"]),
                self.on_plot_fatbands_btn
            )

            d["EffMass"] = self.get_effmass_view()

        else:
            raise ValueError("Neither a IBZ nor k-path!")

        d["Structure"] = self.get_structure_view()
        d["NcFile"] = self.ncfile.get_ncfile_view()

        if as_dict: return d

        return self.get_template_from_tabs(d, template=kwargs.get("template", None))


#class FatbandsRobotPanel(PanelWithEbandsRobot):
#    """
#    A Panel to interoperate with multiple GSR files.
#    """
#
#    gsr_dataframe_btn = pnw.Button(name="Compute", button_type='primary')
#
#    def __init__(self, robot, **params):
#        super().__init__(**params)
#        self.robot = robot
#
#    @param.depends("gsr_dataframe_btn.clicks")
#    def on_gsr_dataframe_btn(self):
#        if self.gsr_dataframe_btn.clicks == 0: return
#        df = self.robot.get_dataframe(with_geo=True)
#        return pn.Column(dfc(df), sizing_mode='stretch_width')
#
#    def get_panel(self):
#        """Return tabs with widgets to interact with the |GsrRobot|."""
#        tabs = pn.Tabs(); app = tabs.append
#        app["Summary"] = self.get_summary_view_for_abiobj(self.robot)
#        app(("e-Bands", pn.Row(self.get_ebands_plotter_widgets(), self.on_ebands_plotter_btn)))
#
#        # Add e-DOS tab only if all ebands have k-sampling.
#        if all(abifile.ebands.kpoints.is_ibz for abifile in self.robot.abifiles):
#            app(("e-DOS", pn.Row(self.get_edos_plotter_widgets(), self.on_edos_plotter_btn)))
#
#        app(("GSR-DataFrame", pn.Row(self.gsr_dataframe_btn, self.on_gsr_dataframe_btn)))
#
#        return tabs



#!/usr/bin/env python
r"""
GWR flow with convergence studies
=================================

This script uses the GWR code to compute the G0W0 corrections
with and without the regularization term in the minimax mesh.
in order to monitor the convergence and stability wrt gwr_ntau.
"""

import os
import sys
import abipy.data as data
import abipy.abilab as abilab

from abipy import flowtk


def build_flow(options):

    from abipy.data.gwr_structures import get_gwr_structure
    symbol = "Si"
    #symbol = "LiF"
    #symbol = "Si"
    #symbol = "C"
    #symbol = "BN"
    #symbol = "MgO"
    structure = get_gwr_structure(symbol)

    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_","flow_")

    from abipy.flowtk.psrepos import get_repo_from_name
    pseudos = get_repo_from_name("ONCVPSP-PBE-SR-PDv0.4").get_pseudos("stringent")

    scf_input = abilab.AbinitInput(structure=structure, pseudos=pseudos)
    scf_input.set_cutoffs_for_accuracy("normal")
    ecut = scf_input["ecut"]
    scf_input.set_scf_nband_semicond()

    # Global variables.
    scf_input.set_vars(
        tolvrs=1e-8,
        paral_kgb=0,
        npfft=1,
    )
    scf_input.set_kmesh(
        #ngkpt=[2, 2, 2],
        ngkpt=[4, 4, 4],
        #ngkpt=[8, 8, 8],
        shiftk=[0.0, 0.0, 0.0], # IMPORTANT: k-grid for GWR must be Gamma-centered.
    )

    # Get max number of PWs.
    dims, _ = scf_input.abiget_dims_spginfo()
    mpw = dims["mpw"]

    flow = flowtk.Flow(workdir=options.workdir)
    small_manager = options.manager.new_with_fixed_mpi_omp(4, 1)

    # GS-SCF run to get the DEN, followed by direct diago to obtain green_nband bands.
    from abipy.flowtk.gwr_works import DirectDiagoWork, GWRSigmaConvWork
    green_nband = -1  # -1 this means full diago
    diago_work = DirectDiagoWork.from_scf_input(scf_input, green_nband)
    diago_work[0].set_manager(small_manager)
    flow.register_work(diago_work)

    # Build template for GWR.
    ecuteps = 12
    gwr_template = scf_input.make_gwr_qprange_input(gwr_ntau=6, nband=int(mpw * 0.9), ecuteps=ecuteps)

    ival = scf_input.num_valence_electrons // 2
    kptgw = [ # k-points in reduced coordinates
        (0.0, 0.0, 0.0),
        (0.5, 0.5, 0.0), # X
        #(0.5    0.000    0.000
    ]

    nkptgw = len(kptgw)
    bdgw = (ival, ival+1) * nkptgw

    gwr_template.set_vars(
        nkptgw=nkptgw,
        kptgw=kptgw,
        bdgw=bdgw,
    )

    gwr_ntau_list = list(range(6, 34, 2))
    gwr_ntau_list = [6, 8]

    # 1) Change the value of one variable:
    varname_values = ("gwr_ntau", gwr_ntau_list)
    # or take the Cartesian product of two or more variables with e.g.:
    #
    #varname_values = [
    #   ("gwr_ntau", gwr_ntau_list),
    #   #("userra", [0.0, 1e-6),    # Compute QP corrections with/without regterm.
    #   #("ecuteps", [4, 6]),
    #]

    # Conpute QP corrections without/with regularization term.
    for userra in [0.0, 1e-6]:
        gwr_template["userra"] = userra
        gwr_work = GWRSigmaConvWork.from_varname_values(
                varname_values, gwr_template, den_node=diago_work[0], wfk_node=diago_work[1])
        flow.register_work(gwr_work)

    return flow


@flowtk.flow_main
def main(options):
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
SCR matrix
==========

This examples shows how to plot the matrix elements of the
inverse dielectric function stored in the SCR file (optdriver 3)
See also `plot_scr.py` for the optical spectrum.
"""
import abipy.data as abidata
from abipy.abilab import abiopen

with abiopen(abidata.ref_file("sio2_SCR.nc")) as ncfile:
    #print(ncfile)
    # The SCR file contains a structure and electron bands in the IBZ.
    # We can thus use the ebands object to plot bands + DOS.
    #ncfile.ebands.plot()

    # Read e^{-1}_{G1, G2}(k, omega)

    kpoint = [0.5, 0, 0]
    em1 = ncfile.reader.read_wggmat(kpoint)
    print(em1)

    gvec1 = [1, 0, 0]
    gvec2 = [-1, 0, 0]

    # Plot matrix element as function of frequency along the real axis
    em1.plot_freq(gvec1=gvec1, gvec2=gvec2, waxis="real")

    # Plot data along the imaginary axis
    # Note that gvectors can also be specifined in terms of their index.
    em1.plot_freq(gvec1=0, gvec2=0, waxis="imag")

    # Plot e^{-1}_{G1, G2} along the imaginary axis.
    em1.plot_gg(wpos="imag")
